from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .model import NousNet


def _logit(p: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    p = p.clamp(eps, 1.0 - eps)
    return torch.log(p) - torch.log1p(-p)


class ThresholdFactBank(nn.Module):
    """
    D*T threshold facts:
      fact_i(x) = sigmoid(kappa_i * (x[feat_i] - th_i))

    Notes
    -----
    - Thresholds are in the same feature space as input X (often standardized).
    - This bank is intended for interpretable, monotone threshold facts.
    """
    def __init__(self, input_dim: int, n_thresh_per_feat: int = 6, init_kappa: float = 2.0) -> None:
        super().__init__()
        self.input_dim = int(input_dim)
        self.n_thresh_per_feat = int(n_thresh_per_feat)
        self.num_facts = self.input_dim * self.n_thresh_per_feat

        feat_idx = torch.arange(self.num_facts) // self.n_thresh_per_feat
        self.register_buffer("feat_idx", feat_idx)

        self.th = nn.Parameter(torch.zeros(self.num_facts))
        self.log_kappa = nn.Parameter(torch.zeros(self.num_facts))
        with torch.no_grad():
            self.th.normal_(0.0, 0.8)
            self.log_kappa.fill_(float(np.log(init_kappa)))

    @torch.no_grad()
    def init_from_data_quantiles(self, X: np.ndarray, q_lo: float = 0.05, q_hi: float = 0.95) -> None:
        Xt = torch.tensor(np.asarray(X), dtype=torch.float32, device=self.th.device)
        D = int(Xt.shape[1])
        for j in range(D):
            mask = (self.feat_idx == j)
            n = int(mask.sum().item())
            if n <= 0:
                continue
            qs = torch.linspace(q_lo, q_hi, steps=n, device=Xt.device)
            self.th[mask] = torch.quantile(Xt[:, j], qs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        xj = x[:, self.feat_idx]
        kappa = torch.exp(self.log_kappa).clamp(0.5, 50.0)
        return torch.sigmoid(kappa * (xj - self.th))


class LogitANDRuleLayer(nn.Module):
    """
    Logit-AND over facts (+ optional negations):
      facts_aug = [facts, 1-facts]
      p = softmax(sel_logits/tau)
      score = logit(facts_aug) @ p^T - bias
      rule = sigmoid(score)

    Returns
    -------
    z : [B, R] rule activations in [0,1]
    p : [R, Fin] selector probabilities
    """
    def __init__(self, n_rules: int, n_facts: int, tau: float = 0.7, use_negations: bool = True) -> None:
        super().__init__()
        self.n_rules = int(n_rules)
        self.n_facts = int(n_facts)
        self.tau = float(tau)
        self.use_negations = bool(use_negations)

        self.fin = self.n_facts * (2 if self.use_negations else 1)
        self.sel_logits = nn.Parameter(torch.zeros(self.n_rules, self.fin))
        self.bias = nn.Parameter(torch.zeros(self.n_rules))
        with torch.no_grad():
            self.sel_logits.normal_(0.0, 0.2)
            self.bias.zero_()

    def selector_probs(self) -> torch.Tensor:
        return torch.softmax(self.sel_logits / max(self.tau, 1e-6), dim=1)

    def forward(self, facts: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        facts_aug = torch.cat([facts, 1.0 - facts], dim=1) if self.use_negations else facts
        p = self.selector_probs()  # [R, Fin]
        score = _logit(facts_aug) @ p.t() - self.bias.unsqueeze(0)
        z = torch.sigmoid(score)
        return z, p


class LowRankInteractionFacts(nn.Module):
    """
    Interaction facts computed from base facts in logit-space:
      h = sigmoid(logit(facts) @ U + b)
    """
    def __init__(self, n_base_facts: int, n_inter: int = 24) -> None:
        super().__init__()
        self.U = nn.Parameter(torch.randn(int(n_base_facts), int(n_inter)) * 0.05)
        self.b = nn.Parameter(torch.zeros(int(n_inter)))

    def forward(self, facts: torch.Tensor) -> torch.Tensor:
        l = _logit(facts)
        return torch.sigmoid(l @ self.U + self.b)


class SoftLogitAND(nn.Module):
    """
    SoftLogitAND classifier:
      - threshold fact bank
      - logit-AND rules
      - linear head on rules -> binary logit
    """
    def __init__(
        self,
        input_dim: int,
        n_rules: int = 64,
        n_thresh_per_feat: int = 6,
        tau: float = 0.7,
        use_negations: bool = True,
    ) -> None:
        super().__init__()
        self.facts = ThresholdFactBank(input_dim, n_thresh_per_feat=n_thresh_per_feat)
        self.rules = LogitANDRuleLayer(n_rules=n_rules, n_facts=self.facts.num_facts, tau=tau, use_negations=use_negations)
        self.head = nn.Linear(n_rules, 1)

    @torch.no_grad()
    def init_from_data(self, X: np.ndarray) -> None:
        self.facts.init_from_data_quantiles(X)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        f = self.facts(x)
        z, _ = self.rules(f)
        return self.head(z).squeeze(-1)


class SoftLogicInteraction(nn.Module):
    """
    SoftLogicInteraction classifier:
      - threshold fact bank
      - low-rank interaction facts
      - logit-AND rules over [facts, interactions]
      - linear head -> binary logit
    """
    def __init__(
        self,
        input_dim: int,
        n_rules: int = 64,
        n_thresh_per_feat: int = 6,
        n_inter: int = 24,
        tau: float = 0.7,
    ) -> None:
        super().__init__()
        self.facts = ThresholdFactBank(input_dim, n_thresh_per_feat=n_thresh_per_feat)
        self.inter = LowRankInteractionFacts(self.facts.num_facts, n_inter=n_inter)
        self.rules = LogitANDRuleLayer(n_rules=n_rules, n_facts=self.facts.num_facts + n_inter, tau=tau, use_negations=True)
        self.head = nn.Linear(n_rules, 1)

    @torch.no_grad()
    def init_from_data(self, X: np.ndarray) -> None:
        self.facts.init_from_data_quantiles(X)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        f = self.facts(x)
        h = self.inter(f)
        fh = torch.cat([f, h], dim=1)
        z, _ = self.rules(fh)
        return self.head(z).squeeze(-1)


def _default_feature_groups(feature_names: Sequence[str]) -> Dict[str, List[int]]:
    """
    Heuristic grouping inspired by WDBC ('mean', 'error', 'worst').
    Falls back to a single group 'all' if no pattern is detected.
    """
    groups: Dict[str, List[int]] = {"mean": [], "error": [], "worst": []}
    for i, n in enumerate(feature_names):
        nl = str(n).lower()
        if nl.startswith("mean "):
            groups["mean"].append(i)
        elif nl.endswith(" error"):
            groups["error"].append(i)
        elif nl.startswith("worst "):
            groups["worst"].append(i)
        else:
            groups.setdefault("other", []).append(i)

    out = {k: v for k, v in groups.items() if len(v) > 0}
    if not out:
        return {"all": list(range(len(feature_names)))}
    return out


def _build_group_fact_indices(feature_names: Sequence[str], n_thresh_per_feat: int) -> Tuple[List[str], List[torch.Tensor]]:
    groups = _default_feature_groups(feature_names)
    keys = list(groups.keys())
    idx_tensors: List[torch.Tensor] = []
    for key in keys:
        feat_idx = groups[key]
        idxs: List[int] = []
        for j in feat_idx:
            idxs.extend(list(range(j * n_thresh_per_feat, (j + 1) * n_thresh_per_feat)))
        idx_tensors.append(torch.tensor(idxs, dtype=torch.long))
    return keys, idx_tensors


class AdditiveThresholdExpert(nn.Module):
    """Additive expert: linear model on facts producing a logit."""
    def __init__(self, n_facts: int) -> None:
        super().__init__()
        self.w = nn.Parameter(torch.randn(int(n_facts), 1) * 0.02)
        self.b = nn.Parameter(torch.zeros(1))

    def forward(self, f: torch.Tensor) -> torch.Tensor:
        return (f @ self.w + self.b).squeeze(-1)


class SegmentMoE(nn.Module):
    """
    SegmentMoE:
      gate(x) computed by group-wise Logit-AND rule layers, combined multiplicatively,
      experts are additive models on full facts, mixed by gate probabilities.
    """
    def __init__(
        self,
        input_dim: int,
        feature_names: Sequence[str],
        n_segments: int = 6,
        n_thresh_per_feat: int = 6,
        tau: float = 0.7,
    ) -> None:
        super().__init__()
        self.feature_names = list(feature_names)
        self.n_segments = int(n_segments)
        self.n_thresh_per_feat = int(n_thresh_per_feat)

        self.facts = ThresholdFactBank(input_dim, n_thresh_per_feat=n_thresh_per_feat)
        self.group_keys, self.group_fact_idx = _build_group_fact_indices(self.feature_names, n_thresh_per_feat)

        self.group_layers = nn.ModuleList([
            LogitANDRuleLayer(n_rules=self.n_segments, n_facts=int(idx.numel()), tau=tau, use_negations=True)
            for idx in self.group_fact_idx
        ])
        self.pi_logits = nn.Parameter(torch.zeros(self.n_segments))
        self.experts = nn.ModuleList([AdditiveThresholdExpert(self.facts.num_facts) for _ in range(self.n_segments)])

    @torch.no_grad()
    def init_from_data(self, X: np.ndarray) -> None:
        self.facts.init_from_data_quantiles(X)

    def gate_probs_from_facts(self, f: torch.Tensor) -> torch.Tensor:
        parts = []
        for idx, layer in zip(self.group_fact_idx, self.group_layers):
            fg = f[:, idx.to(f.device)]
            zg, _ = layer(fg)               # [B, K]
            parts.append(zg.unsqueeze(-1))  # [B, K, 1]
        A = torch.cat(parts, dim=-1)        # [B, K, G]
        m = A.prod(dim=-1)                  # [B, K]
        pi = torch.softmax(self.pi_logits, dim=0).unsqueeze(0)  # [1, K]
        w = (m * pi).clamp_min(1e-12)
        return w / (w.sum(dim=1, keepdim=True) + 1e-12)         # [B, K]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        f = self.facts(x)
        gprob = self.gate_probs_from_facts(f)  # [B, K]
        expert_logits = torch.stack([ex(f) for ex in self.experts], dim=1)  # [B, K]
        return (gprob * expert_logits).sum(dim=1)  # [B]


class SoftLogicExpert(nn.Module):
    """Soft logic expert used inside HierarchicalMoE."""
    def __init__(self, n_facts: int, n_rules: int = 24, tau: float = 0.7) -> None:
        super().__init__()
        self.rules = LogitANDRuleLayer(n_rules=n_rules, n_facts=n_facts, tau=tau, use_negations=True)
        self.head = nn.Linear(n_rules, 1)

    def forward(self, f: torch.Tensor) -> torch.Tensor:
        z, _ = self.rules(f)
        return self.head(z).squeeze(-1)


class HierarchicalMoE(nn.Module):
    """
    HierarchicalMoE:
      - gate is SegmentMoE-like (group-wise rule gates),
      - experts are SoftLogicExpert models.
    """
    def __init__(
        self,
        input_dim: int,
        feature_names: Sequence[str],
        n_segments: int = 5,
        n_rules_expert: int = 24,
        n_thresh_per_feat: int = 6,
        tau: float = 0.7,
    ) -> None:
        super().__init__()
        self.feature_names = list(feature_names)
        self.n_segments = int(n_segments)
        self.n_thresh_per_feat = int(n_thresh_per_feat)

        self.facts = ThresholdFactBank(input_dim, n_thresh_per_feat=n_thresh_per_feat)
        self.group_keys, self.group_fact_idx = _build_group_fact_indices(self.feature_names, n_thresh_per_feat)

        self.group_layers = nn.ModuleList([
            LogitANDRuleLayer(n_rules=self.n_segments, n_facts=int(idx.numel()), tau=tau, use_negations=True)
            for idx in self.group_fact_idx
        ])
        self.pi_logits = nn.Parameter(torch.zeros(self.n_segments))
        self.experts = nn.ModuleList([
            SoftLogicExpert(self.facts.num_facts, n_rules=n_rules_expert, tau=tau) for _ in range(self.n_segments)
        ])

    @torch.no_grad()
    def init_from_data(self, X: np.ndarray) -> None:
        self.facts.init_from_data_quantiles(X)

    def gate_probs_from_facts(self, f: torch.Tensor) -> torch.Tensor:
        parts = []
        for idx, layer in zip(self.group_fact_idx, self.group_layers):
            fg = f[:, idx.to(f.device)]
            zg, _ = layer(fg)
            parts.append(zg.unsqueeze(-1))
        A = torch.cat(parts, dim=-1)
        m = A.prod(dim=-1)
        pi = torch.softmax(self.pi_logits, dim=0).unsqueeze(0)
        w = (m * pi).clamp_min(1e-12)
        return w / (w.sum(dim=1, keepdim=True) + 1e-12)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        f = self.facts(x)
        gprob = self.gate_probs_from_facts(f)  # [B, K]
        expert_logits = torch.stack([ex(f) for ex in self.experts], dim=1)  # [B, K]
        return (gprob * expert_logits).sum(dim=1)


class FixedWidthThresholdFacts(nn.Module):
    """
    Exactly num_facts threshold facts (for NousFamilies backbone):
      fact_i(x) = sigmoid(kappa_i * (x[feat_i] - th_i))

    The feature index for fact i is (i % input_dim).
    """
    def __init__(self, input_dim: int, num_facts: int, init_kappa: float = 2.0) -> None:
        super().__init__()
        self.input_dim = int(input_dim)
        self.num_facts = int(num_facts)

        feat_idx = torch.arange(self.num_facts) % self.input_dim
        self.register_buffer("feat_idx", feat_idx)

        self.th = nn.Parameter(torch.zeros(self.num_facts))
        self.log_kappa = nn.Parameter(torch.zeros(self.num_facts))
        with torch.no_grad():
            self.th.normal_(0.0, 0.8)
            self.log_kappa.fill_(float(np.log(init_kappa)))

    @torch.no_grad()
    def init_from_data_quantiles(self, X: np.ndarray, q_lo: float = 0.05, q_hi: float = 0.95) -> None:
        Xt = torch.tensor(np.asarray(X), dtype=torch.float32, device=self.th.device)
        D = int(Xt.shape[1])
        for j in range(D):
            mask = (self.feat_idx == j)
            n = int(mask.sum().item())
            if n <= 0:
                continue
            qs = torch.linspace(q_lo, q_hi, steps=n, device=Xt.device)
            self.th[mask] = torch.quantile(Xt[:, j], qs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        xj = x[:, self.feat_idx]
        kappa = torch.exp(self.log_kappa).clamp(0.5, 50.0)
        return torch.sigmoid(kappa * (xj - self.th))


class FamiliesHead(nn.Module):
    """
    Families head:
      - soft assignment A: [R, F] via softmax
      - family activations: fam = (Z @ A) / sum(A)  (per-family normalization)
      - output logit: fam @ W + b
    """
    def __init__(self, n_rules: int, n_families: int, tau: float = 0.7) -> None:
        super().__init__()
        self.n_rules = int(n_rules)
        self.n_families = int(n_families)
        self.tau = float(tau)

        self.A_logits = nn.Parameter(torch.zeros(self.n_rules, self.n_families))
        self.W = nn.Parameter(torch.randn(self.n_families, 1) * 0.05)
        self.b = nn.Parameter(torch.zeros(1))
        with torch.no_grad():
            self.A_logits.normal_(0.0, 0.2)

    def assignment(self) -> torch.Tensor:
        return torch.softmax(self.A_logits / max(self.tau, 1e-6), dim=1)

    def forward(self, Z: torch.Tensor) -> torch.Tensor:
        A = self.assignment()                       # [R, F]
        denom = A.sum(dim=0, keepdim=True) + 1e-12
        fam = (Z @ A) / denom                       # [B, F]
        return (fam @ self.W + self.b).squeeze(-1)  # [B]


class NousFamilies(nn.Module):
    """
    NousFamilies:
      - NousNet backbone produces internal rule activations (gated_activations),
      - FamiliesHead aggregates all rules into stable "families",
      - Output is a single binary logit.

    Notes
    -----
    By default uses NousNet 'soft_fact' layers. We replace backbone.fact with FixedWidthThresholdFacts
    to align with threshold-based interpretable inputs.
    """
    def __init__(
        self,
        input_dim: int,
        feature_names: Sequence[str],
        num_facts: int = 64,
        rules_per_layer: Sequence[int] = (48, 24),
        n_families: int = 12,
    ) -> None:
        super().__init__()
        self.feature_names = list(feature_names)
        self.num_facts = int(num_facts)
        self.rules_per_layer = tuple(int(x) for x in rules_per_layer if int(x) > 0)
        self.n_rules_total = int(sum(self.rules_per_layer))
        self.n_families = int(n_families)

        self.backbone = NousNet(
            input_dim=input_dim,
            num_outputs=1,
            task_type="classification",
            feature_names=feature_names,
            num_facts=self.num_facts,
            rules_per_layer=self.rules_per_layer,
            rule_selection_method="soft_fact",
            use_calibrators=False,
            use_prototypes=False,
        )
        # Replace beta facts with threshold facts (keeps output shape [B, num_facts])
        self.backbone.fact = FixedWidthThresholdFacts(input_dim=input_dim, num_facts=self.num_facts)

        # Freeze backbone head (we do not use it for the final logit)
        if hasattr(self.backbone, "head") and self.backbone.head is not None:
            for p in self.backbone.head.parameters():
                p.requires_grad = False

        self.head = FamiliesHead(self.n_rules_total, self.n_families, tau=0.7)

    @torch.no_grad()
    def init_from_data(self, X: np.ndarray) -> None:
        self.backbone.fact.init_from_data_quantiles(X)

    def extract_rule_matrix(self, x: torch.Tensor, detach_internals: bool = False) -> torch.Tensor:
        """
        Return concatenated gated rule activations across all blocks: [B, sum(rules_per_layer)].
        """
        _, internals = self.backbone(x, return_internals=True, detach_internals=detach_internals)
        acts: List[torch.Tensor] = []
        for b_idx in range(len(self.backbone.blocks)):
            blk = internals.get(f"block_{b_idx}", {})
            a = blk.get("gated_activations", None)
            if a is None:
                raise RuntimeError("Expected 'gated_activations' in backbone internals.")
            acts.append(a)
        return torch.cat(acts, dim=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        Z = self.extract_rule_matrix(x, detach_internals=False)
        return self.head(Z)


import numpy as np
import pandas as pd
import torch
from sklearn.preprocessing import StandardScaler

from nous import SoftLogitAND
from nous.explain.softlogitand_posthoc import (
    SoftLogitANDPosthocExplainer,
    hard_cap_topk_rules,
)


def _make_data(n=512, d=20, seed=0):
    rng = np.random.RandomState(seed)
    X = rng.randn(n, d).astype(np.float32)
    # Simple synthetic labels: noisy linear rule
    w = rng.randn(d).astype(np.float32)
    logits = (X @ w) * 0.3
    p = 1.0 / (1.0 + np.exp(-logits))
    y = (rng.rand(n) < p).astype(np.float32)
    return X, y


def test_softlogitand_posthoc_smoke_and_strategies():
    X_raw, y = _make_data(n=800, d=30, seed=1)
    feature_names = [f"x{i:03d}" for i in range(X_raw.shape[1])]

    # Scale for SoftLogitAND training/inference
    scaler = StandardScaler().fit(X_raw[:600])
    X_scaled = scaler.transform(X_raw).astype(np.float32)

    # Model (no training needed for smoke; init_from_data requires X)
    model = SoftLogitAND(input_dim=X_scaled.shape[1], n_rules=64, n_thresh_per_feat=4, tau=0.7, use_negations=True)
    model.init_from_data(X_scaled[:600])

    expl = SoftLogitANDPosthocExplainer(
        model=model,
        feature_names=feature_names,
        x_scaler=scaler,
        k_rules=10,
        k_literals=3,
        n_clusters=8,
        use_surrogate=True,
        surrogate_alpha=1e-3,
        z_threshold=0.5,
    )

    # Fit posthoc on reference distribution
    expl.fit_posthoc(X_ref_scaled=X_scaled[:600])

    # Fit rule stats + add mean abs contrib
    expl.fit_rule_stats(
        X_train_scaled=X_scaled[:600],
        y_train01=y[:600],
        X_test_scaled=X_scaled[600:],
        y_test01=y[600:],
    )
    expl.add_mean_abs_contrib_to_rule_stats(X_train_scaled=X_scaled[:600], X_test_scaled=X_scaled[600:])

    # Local report
    x0 = X_raw[0]
    rep = expl.report(x0, y_true01=float(y[0]), x_is_scaled=False, top_clusters=6)
    assert isinstance(rep.markdown, str) and len(rep.markdown) > 10
    assert isinstance(rep.tables, dict) and len(rep.tables) > 0

    # Ensure strategy tables exist (surrogate may exist because use_surrogate=True)
    expected_keys = {
        "mmr_plus_hardcap_rules",
        "vanilla_rules",
        "mmr_topL_rules",
        "mmr_feature_mass_rules",
        "hardcap_rules",
        "cluster_themes",
        "surrogate_top_rules",
    }
    # Surrogate may fail to fit in some extreme cases; allow missing surrogate key
    missing = expected_keys - set(rep.tables.keys())
    assert missing.issubset({"surrogate_top_rules"}), f"Missing unexpected tables: {missing}"

    # Cluster themes table sanity
    df_ct = rep.tables["cluster_themes"]
    assert isinstance(df_ct, pd.DataFrame)
    assert len(df_ct) > 0
    assert "cluster" in df_ct.columns
    assert "cluster_contrib" in df_ct.columns

    # Fidelity table sanity
    df_fid = expl.fidelity_table_for_x(x0, x_is_scaled=False)
    assert isinstance(df_fid, pd.DataFrame)
    assert len(df_fid) > 0
    assert "coverage_abs_contrib" in df_fid.columns

    # Global tables
    df_rules = expl.global_rule_importance(split="train", topn=10, sort_by="mean_abs_contrib")
    assert isinstance(df_rules, pd.DataFrame) and len(df_rules) > 0

    df_clusters = expl.global_cluster_importance(split="train")
    assert isinstance(df_clusters, pd.DataFrame) and len(df_clusters) > 0

    df_feats = expl.global_feature_importance_mass_weighted(split="train", topn=10)
    assert isinstance(df_feats, pd.DataFrame) and len(df_feats) > 0


def test_hardcap_respects_cap_per_feature():
    rng = np.random.RandomState(0)
    R = 50
    contrib_abs = np.abs(rng.randn(R))
    top_feature_ids = rng.randint(0, 5, size=R)  # 5 dominant features
    k = 20
    cap = 2

    chosen = hard_cap_topk_rules(contrib_abs, top_feature_ids, k=k, cap_per_feature=cap)
    assert len(chosen) <= k

    # Verify cap is respected
    counts = {}
    for r in chosen:
        f = int(top_feature_ids[r])
        counts[f] = counts.get(f, 0) + 1
    assert all(v <= cap for v in counts.values())


from __future__ import annotations

from typing import Any, Dict, List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
import torch

from ..zoo import (
    HierarchicalMoE,
    SegmentMoE,
    SoftLogitAND,
    SoftLogicInteraction,
    ThresholdFactBank,
    NousFamilies,
)


def _invert_standard_scaler(th_scaled: float, feat_idx: int, scaler: Any) -> float:
    # Works with sklearn.preprocessing.StandardScaler
    return float(th_scaled * float(scaler.scale_[feat_idx]) + float(scaler.mean_[feat_idx]))


def describe_threshold_fact(
    fact_bank: ThresholdFactBank,
    fact_index: int,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
    negated: bool = False,
) -> str:
    """
    Render a single threshold fact into a readable condition.
    """
    feat = int(fact_bank.feat_idx[fact_index].item())
    th_scaled = float(fact_bank.th[fact_index].detach().cpu().item())
    if scaler is not None:
        th = _invert_standard_scaler(th_scaled, feat, scaler)
    else:
        th = th_scaled
    name = str(feature_names[feat])
    op = "<=" if negated else ">"
    return f"{name} {op} {th:.4g}"


def softlogitand_top_conditions(
    fact_bank: ThresholdFactBank,
    selector_probs: torch.Tensor,  # [R, Fin]
    rule_idx: int,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
    top_k: int = 3,
) -> List[str]:
    """
    Pick top-k fact conditions by selector probability for a given rule.
    """
    p = selector_probs[int(rule_idx)].detach().cpu().numpy()
    fin = int(p.shape[0])
    F0 = int(fact_bank.num_facts)

    items: List[Tuple[float, str]] = []
    for j in range(fin):
        prob = float(p[j])
        if fin == 2 * F0:
            base = j % F0
            neg = (j >= F0)
            s = describe_threshold_fact(fact_bank, base, feature_names, scaler=scaler, negated=neg)
        else:
            s = describe_threshold_fact(fact_bank, j, feature_names, scaler=scaler, negated=False)
        items.append((prob, s))
    items.sort(key=lambda t: -t[0])
    return [f"{s} (p={prob:.3f})" for prob, s in items[: int(top_k)]]


def softlogitand_global_rules_df(
    model: SoftLogitAND,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
    top_rules: int = 10,
    top_facts_per_rule: int = 3,
) -> pd.DataFrame:
    """
    Global view: rank rules by |head weight| and show their top conditions.
    """
    model.eval()
    with torch.no_grad():
        dummy = torch.zeros(1, model.facts.num_facts, device=next(model.parameters()).device)
        _, P = model.rules(dummy)  # [R, 2F]
        w = model.head.weight.view(-1).detach().cpu().numpy()

    idx = np.argsort(-np.abs(w))[: int(top_rules)]
    rows = []
    for r in idx:
        conds = softlogitand_top_conditions(model.facts, P, int(r), feature_names, scaler=scaler, top_k=top_facts_per_rule)
        rows.append({
            "rule": int(r),
            "weight": float(w[r]),
            "abs_weight": float(abs(w[r])),
            "rule_text": " AND ".join([c.split(" (p=")[0] for c in conds]),
            "top_conditions": " | ".join(conds),
        })
    return pd.DataFrame(rows)


def softlogitand_local_contrib_df(
    model: SoftLogitAND,
    x: np.ndarray,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
    top_rules: int = 10,
    top_facts_per_rule: int = 3,
) -> Tuple[pd.DataFrame, Dict[str, float]]:
    """
    Local view: compute per-rule contributions to the final logit: contrib = z_r * w_r.
    """
    model.eval()
    dev = next(model.parameters()).device
    xt = torch.tensor(np.asarray(x, dtype=np.float32)[None, :], device=dev)

    with torch.no_grad():
        f = model.facts(xt)
        z, P = model.rules(f)  # [1,R], [R,2F]
        w = model.head.weight.view(-1)
        b = model.head.bias.view(-1)[0] if model.head.bias is not None else torch.tensor(0.0, device=dev)
        contrib = (z.view(-1) * w).detach().cpu().numpy()
        z_np = z.view(-1).detach().cpu().numpy()
        w_np = w.detach().cpu().numpy()
        logit = float((z @ w.view(-1, 1) + b).view(-1).item())
        prob = float(torch.sigmoid(torch.tensor(logit)).item())

    idx = np.argsort(-np.abs(contrib))[: int(top_rules)]
    rows = []
    for r in idx:
        conds = softlogitand_top_conditions(model.facts, P, int(r), feature_names, scaler=scaler, top_k=top_facts_per_rule)
        rows.append({
            "rule": int(r),
            "activation": float(z_np[r]),
            "weight": float(w_np[r]),
            "contribution_to_logit": float(contrib[r]),
            "rule_text": " AND ".join([c.split(" (p=")[0] for c in conds]),
        })
    meta = {"logit": logit, "prob_pos": prob}
    return pd.DataFrame(rows), meta


@torch.no_grad()
def moe_gate_probs(model: Any, X: np.ndarray) -> np.ndarray:
    """
    Get gate probabilities for SegmentMoE / HierarchicalMoE.
    """
    model.eval()
    dev = next(model.parameters()).device
    xt = torch.tensor(np.asarray(X, dtype=np.float32), device=dev)
    f = model.facts(xt)
    g = model.gate_probs_from_facts(f)
    return g.detach().cpu().numpy()


def moe_gate_summary_df(model: Any, X_ref: np.ndarray) -> pd.DataFrame:
    """
    Global view: average segment probabilities.
    """
    gp = moe_gate_probs(model, X_ref)
    avg = gp.mean(axis=0)
    return pd.DataFrame({"segment": np.arange(len(avg)), "avg_gate_prob": avg}).sort_values(
        "avg_gate_prob", ascending=False
    ).reset_index(drop=True)


def segmentmoe_local_explain_df(
    model: SegmentMoE,
    x: np.ndarray,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
    top_segments: int = 3,
    top_facts: int = 8,
) -> Tuple[pd.DataFrame, Dict[str, float]]:
    """
    Local SegmentMoE explanation:
      - top segments by gate prob,
      - expert logit,
      - segment contribution to final logit,
      - top additive fact contributions inside the expert.
    """
    model.eval()
    dev = next(model.parameters()).device
    xt = torch.tensor(np.asarray(x, dtype=np.float32)[None, :], device=dev)

    with torch.no_grad():
        f = model.facts(xt)
        g = moe_gate_probs(model, np.asarray(x, dtype=np.float32)[None, :])[0]
        expert_logits = torch.stack([ex(f) for ex in model.experts], dim=1).view(-1).detach().cpu().numpy()
        logit = float((g * expert_logits).sum())
        prob = float(1.0 / (1.0 + np.exp(-logit)))

    seg_idx = np.argsort(-g)[: int(top_segments)]
    rows = []
    for s in seg_idx:
        ex = model.experts[int(s)]
        with torch.no_grad():
            w = ex.w.view(-1).detach().cpu().numpy()
            f_np = f.view(-1).detach().cpu().numpy()
            contrib = f_np * w

        top = np.argsort(-np.abs(contrib))[: int(top_facts)]
        items = []
        for fi in top:
            # SegmentMoE uses ThresholdFactBank facts (not negations)
            text = describe_threshold_fact(model.facts, int(fi), feature_names, scaler=scaler, negated=False)
            items.append(f"{text}: {contrib[fi]:+.3f}")

        rows.append({
            "segment": int(s),
            "gate_prob": float(g[s]),
            "expert_logit": float(expert_logits[s]),
            "segment_contribution_to_logit": float(g[s] * expert_logits[s]),
            "top_fact_contribs": " | ".join(items),
        })
    meta = {"logit": logit, "prob_pos": prob}
    return pd.DataFrame(rows), meta


def hiermoe_local_explain_df(
    model: HierarchicalMoE,
    x: np.ndarray,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
    top_segments: int = 3,
    top_rules: int = 6,
    top_facts_per_rule: int = 3,
) -> Tuple[pd.DataFrame, Dict[str, float]]:
    """
    Local HierarchicalMoE explanation:
      - top segments by gate prob,
      - expert logit + segment contribution,
      - inside each selected expert: top contributing rules (approx) with their top conditions.
    """
    model.eval()
    dev = next(model.parameters()).device
    xt = torch.tensor(np.asarray(x, dtype=np.float32)[None, :], device=dev)

    with torch.no_grad():
        f = model.facts(xt)
        g = moe_gate_probs(model, np.asarray(x, dtype=np.float32)[None, :])[0]
        expert_logits = torch.stack([ex(f) for ex in model.experts], dim=1).view(-1).detach().cpu().numpy()
        logit = float((g * expert_logits).sum())
        prob = float(1.0 / (1.0 + np.exp(-logit)))

    seg_idx = np.argsort(-g)[: int(top_segments)]
    rows = []
    for s in seg_idx:
        ex = model.experts[int(s)]
        with torch.no_grad():
            z, P = ex.rules(f)  # [1,R], [R,2F]
            w = ex.head.weight.view(-1).detach().cpu().numpy()
            z_np = z.view(-1).detach().cpu().numpy()
            contrib = z_np * w

        ridx = np.argsort(-np.abs(contrib))[: int(top_rules)]
        rule_texts = []
        for r in ridx:
            conds = softlogitand_top_conditions(model.facts, P, int(r), feature_names, scaler=scaler, top_k=top_facts_per_rule)
            rule_texts.append(f"[r{int(r)}] {contrib[r]:+.3f}: " + " AND ".join([c.split(" (p=")[0] for c in conds]))

        rows.append({
            "segment": int(s),
            "gate_prob": float(g[s]),
            "expert_logit": float(expert_logits[s]),
            "segment_contribution_to_logit": float(g[s] * expert_logits[s]),
            "top_expert_rules": " || ".join(rule_texts),
        })
    meta = {"logit": logit, "prob_pos": prob}
    return pd.DataFrame(rows), meta


def nousfamilies_global_summary_df(model: NousFamilies, top_families: int = 10) -> pd.DataFrame:
    """
    Global view for NousFamilies:
      - family weights,
      - "rule mass" assigned to each family (sum over rules of assignment probability).
    """
    model.eval()
    with torch.no_grad():
        W = model.head.W.view(-1).detach().cpu().numpy()
        A = model.head.assignment().detach().cpu().numpy()  # [R, F]
    fam_mass = A.sum(axis=0)

    df = pd.DataFrame({
        "family": np.arange(model.n_families),
        "weight_W": W,
        "abs_weight": np.abs(W),
        "assigned_rule_mass": fam_mass,
    })
    return df.sort_values("abs_weight", ascending=False).head(int(top_families)).reset_index(drop=True)


def nousfamilies_local_contrib_df(
    model: NousFamilies,
    x: np.ndarray,
    top_families: int = 8,
) -> Tuple[pd.DataFrame, Dict[str, float]]:
    """
    Local view for NousFamilies:
      - family activations,
      - family contributions to the final logit.
    """
    model.eval()
    dev = next(model.parameters()).device
    xt = torch.tensor(np.asarray(x, dtype=np.float32)[None, :], device=dev)

    with torch.no_grad():
        Z = model.extract_rule_matrix(xt, detach_internals=True)  # [1,R]
        A = model.head.assignment()                                # [R,F]
        denom = A.sum(dim=0, keepdim=True) + 1e-12
        fam = (Z @ A) / denom                                      # [1,F]
        W = model.head.W.view(-1)
        contrib = (fam.view(-1) * W).detach().cpu().numpy()

        logit = float(model(xt).view(-1).item())
        prob = float(1.0 / (1.0 + np.exp(-logit)))

    idx = np.argsort(-np.abs(contrib))[: int(top_families)]
    df = pd.DataFrame({
        "family": idx.astype(int),
        "family_activation": fam.view(-1).detach().cpu().numpy()[idx],
        "family_weight_W": W.detach().cpu().numpy()[idx],
        "family_contribution_to_logit": contrib[idx],
    }).reset_index(drop=True)

    meta = {"logit": logit, "prob_pos": prob}
    return df, meta


import numpy as np
import torch

from nous import SoftLogitAND, SoftLogicInteraction, SegmentMoE, HierarchicalMoE, NousFamilies


def test_zoo_models_forward_and_backward_smoke():
    torch.manual_seed(0)
    np.random.seed(0)

    D = 6
    B = 8
    feature_names = ["mean a", "mean b", "worst c", "x error", "other1", "other2"]

    X = np.random.randn(64, D).astype(np.float32)
    xb = torch.tensor(np.random.randn(B, D).astype(np.float32))
    yb = torch.randint(0, 2, (B,), dtype=torch.float32)

    models = [
        SoftLogitAND(D, n_rules=16, n_thresh_per_feat=4),
        SoftLogicInteraction(D, n_rules=16, n_thresh_per_feat=4, n_inter=8),
        SegmentMoE(D, feature_names=feature_names, n_segments=4, n_thresh_per_feat=4),
        HierarchicalMoE(D, feature_names=feature_names, n_segments=4, n_rules_expert=8, n_thresh_per_feat=4),
        NousFamilies(D, feature_names=feature_names, num_facts=16, rules_per_layer=(8, 6), n_families=5),
    ]

    for m in models:
        if hasattr(m, "init_from_data"):
            m.init_from_data(X)

        out = m(xb)
        assert out.shape == (B,)

        loss = torch.nn.BCEWithLogitsLoss()(out, yb)
        loss.backward()


from __future__ import annotations
import torch
import torch.nn.functional as F
from torch.optim import Optimizer
from typing import Dict, Iterable, Optional, Tuple
import numpy as np
from .base import BaseOptimizer

class MirrorRAdam(BaseOptimizer):
    """RAdam optimizer with mirror descent updates for gate parameters in KL geometry.
    This optimizer applies standard RAdam updates to non-gate parameters,
    while using a mirror descent step in the KL geometry for gate parameters.
    This approach stabilizes optimization of softmax-based gating mechanisms
    used in rule selection layers.
    Args:
        params: Iterable of parameters to optimize
        model: The model being optimized (required for parameter categorization)
        lr: Learning rate for non-gate parameters (default: 1e-3)
        lr_gate: Learning rate for gate parameters (default: 1e-2)
        betas: Coefficients used for computing running averages of gradient and squared gradient (default: (0.9, 0.999))
        eps: Term added to the denominator to improve numerical stability (default: 1e-8)
        weight_decay: Weight decay coefficient (default: 0)
    """
    def __init__(
        self,
        params: Iterable[torch.nn.parameter.Parameter],
        model: torch.nn.Module,
        lr: float = 1e-3,
        lr_gate: float = 1e-2,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0.0,
    ):
        if lr <= 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if lr_gate <= 0.0:
            raise ValueError(f"Invalid gate learning rate: {lr_gate}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta1 parameter: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta2 parameter: {betas[1]}")
        if eps <= 0.0:
            raise ValueError(f"Invalid epsilon value: {eps}")
        defaults = dict(
            lr=lr,
            lr_gate=lr_gate,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            step=0
        )
        super().__init__(params, defaults, model=model)
        # Create separate parameter groups for non-gate parameters
        self.param_groups = [
            {
                'params': self.nongate_params,
                'lr': lr,
                'weight_decay': weight_decay,
                'betas': betas,
                'eps': eps
            }
        ]

    def _radam_rectifier(self, t: int, beta2: float):
        """RAdam variance rectification term."""
        rho_inf = 2.0 / (1.0 - beta2) - 1.0
        rho_t = rho_inf - 2.0 * (beta2 ** t) / (1.0 - beta2 ** t)
        if rho_t > 4.0:
            rt = np.sqrt(((rho_t - 4.0) * (rho_t - 2.0) * rho_inf) / ((rho_inf - 4.0) * (rho_inf - 2.0) * rho_t))
            return float(rt), True
        return 1.0, False

    def step(self, closure=None):
        """Performs a single optimization step."""
        loss = None
        if closure is not None:
            loss = closure()

        # Update step count
        for group in self.param_groups:
            group['step'] = group.get('step', 0) + 1

        # Get non-gate parameter group
        non_gate_group = self.param_groups[0]
        params = non_gate_group['params']
        lr = non_gate_group['lr']
        weight_decay = non_gate_group['weight_decay']
        beta1, beta2 = non_gate_group['betas']
        eps = non_gate_group['eps']
        step = non_gate_group['step']

        # Update non-gate parameters with RAdam
        for p in params:
            if p.grad is None:
                continue
            grad = p.grad.data
            state = self.state[p]

            # State initialization
            if len(state) == 0:
                state['step'] = 0
                state['exp_avg'] = torch.zeros_like(p.data)
                state['exp_avg_sq'] = torch.zeros_like(p.data)
            state['step'] += 1
            step = state['step']
            exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']

            # Decay the first and second moment running average coefficient
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

            # Compute bias-corrected moments
            bias_correction1 = 1 - beta1 ** step
            bias_correction2 = 1 - beta2 ** step

            # Compute rectification term
            rt, rectified = self._radam_rectifier(step, beta2)
            if rectified:
                # Compute variance rectification
                step_size = lr * rt * np.sqrt(bias_correction2) / bias_correction1
                # Denominator with bias correction
                denom = exp_avg_sq.sqrt() / np.sqrt(bias_correction2) + eps
                # Update parameters
                if weight_decay != 0:
                    p.data.add_(p.data, alpha=-weight_decay * lr)
                p.data.addcdiv_(exp_avg, denom, value=-step_size)
            else:
                # Standard Adam update
                step_size = lr / bias_correction1
                denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
                if weight_decay != 0:
                    p.data.add_(p.data, alpha=-weight_decay * lr)
                p.data.addcdiv_(exp_avg, denom, value=-step_size)

        # Handle gate parameters with mirror descent in KL geometry
        gate_lr = self.defaults['lr_gate']
        for name, p in self.model.named_parameters():
            if p.grad is None or not any(p is param for param in self.gate_params):
                continue

            grad = p.grad.data
            # Optimizer state keyed by parameter tensor (PyTorch convention)
            state = self.state[p]
            if len(state) == 0:
                state['exp_avg'] = torch.zeros_like(p.data)
                state['exp_avg_sq'] = torch.zeros_like(p.data)
                state['step'] = 0
            state['step'] += 1
            step = state['step']

            exp_avg = state['exp_avg']
            exp_avg_sq = state['exp_avg_sq']

            # Update moving averages
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

            # Bias correction
            bias_correction1 = 1 - beta1 ** step
            bias_correction2 = 1 - beta2 ** step
            rt, rectified = self._radam_rectifier(step, beta2)

            # Compute step
            if rectified:
                step_size = gate_lr * rt * np.sqrt(bias_correction2) / bias_correction1
                denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
                step = (exp_avg / denom) * step_size
            else:
                step_size = gate_lr / bias_correction1
                denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
                step = (exp_avg / denom) * step_size

            # Mirror descent step in KL geometry
            if p.dim() == 2:
                # For softmax gates (row-wise updates)
                P_old = F.softmax(p.data, dim=1)
                P_new = P_old * torch.exp(-step)
                P_new = P_new / (P_new.sum(dim=1, keepdim=True) + 1e-8)
                logits_new = torch.log(torch.clamp(P_new, 1e-8, 1.0))
                logits_new -= logits_new.mean(dim=1, keepdim=True)
                p.data.copy_(logits_new)
            else:
                # For scalar gates
                p.data.add_(-step)

            # Zero the gradient after update
            if p.grad is not None:
                p.grad.zero_()

        return loss


from __future__ import annotations
import torch
from torch.optim import Optimizer
from typing import Dict, Iterable, Optional, Tuple

class BaseOptimizer(Optimizer):
    """Base class for Nous-specific optimizers with parameter grouping."""
    
    def __init__(self, params, defaults, model=None):
        super().__init__(params, defaults)
        self.gate_params = []
        self.nongate_params = []
        self.model = model
        self._categorize_parameters()
    
    def _categorize_parameters(self):
        """Categorize parameters into gate and non-gate groups."""
        self.gate_params = []
        self.nongate_params = []
        
        if self.model is not None:
            # Use model's named_parameters to get parameter names
            for name, param in self.model.named_parameters():
                if any(p is param for p in self.param_groups[0]['params']):
                    if self._is_gate_param(name):
                        self.gate_params.append(param)
                    else:
                        self.nongate_params.append(param)
        else:
            # Fallback: check parameter names if available in state
            for group in self.param_groups:
                for p in group['params']:
                    name = getattr(p, '_name', None) or getattr(p, 'name', None)
                    if name and self._is_gate_param(name):
                        self.gate_params.append(p)
                    else:
                        self.nongate_params.append(p)
    
    def _is_gate_param(self, name: str) -> bool:
        """Check if parameter is a gate parameter."""
        if not name:
            return False
        return ("fact_logits" in name) or ("aggregator_logits" in name) or ("rule_strength_raw" in name)
    
    def step(self, closure=None):
        """Performs a single optimization step."""
        raise NotImplementedError("Subclasses must implement step()")



from .mirror_radam import MirrorRAdam
from .beta_ng import BetaNaturalGradient
from .extragrad_trust import ExtragradRAdamTrust
from .sam_gates import SharpnessAwareGates
from .factory import create_optimizer

__all__ = [
    "MirrorRAdam",
    "BetaNaturalGradient",
    "ExtragradRAdamTrust",
    "SharpnessAwareGates",
    "create_optimizer",
]


from __future__ import annotations
import torch
import torch.nn.functional as F
import numpy as np
from torch.optim import Optimizer
from typing import Dict, Iterable, Optional, Tuple

from .base import BaseOptimizer
from ..model import NousNet

class BetaNaturalGradient(BaseOptimizer):
    """Stabilized β-natural gradient preconditioner for BetaFactLayer parameters.
    
    Uses diagonal Fisher approximation with EMA smoothing and trust-ratio clipping
    to stabilize optimization of β-facts parameters.
    
    Args:
        params: Iterable of parameters to optimize
        model: The model being optimized (required for parameter categorization)
        lr: Learning rate (default: 1e-3)
        ema_beta: EMA decay for Fisher approximation (default: 0.9)
        eps: Small value for numerical stability (default: 1e-6)
        trust_clip: Maximum trust ratio for step size (default: 10.0)
        weight_decay: Weight decay coefficient (default: 0)
    """
    
    def __init__(
        self,
        params: Iterable[torch.nn.parameter.Parameter],
        model: torch.nn.Module,
        lr: float = 1e-3,
        ema_beta: float = 0.9,
        eps: float = 1e-6,
        trust_clip: float = 10.0,
        weight_decay: float = 0.0,
    ):
        if lr <= 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= ema_beta < 1.0:
            raise ValueError(f"Invalid EMA beta parameter: {ema_beta}")
        if eps <= 0.0:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if trust_clip <= 0.0:
            raise ValueError(f"Invalid trust clip value: {trust_clip}")
        
        defaults = dict(
            lr=lr,
            ema_beta=ema_beta,
            eps=eps,
            trust_clip=trust_clip,
            weight_decay=weight_decay,
            step=0
        )
        super().__init__(params, defaults, model=model)
        
        # Fisher approximation state
        self.ema_state: Dict[str, torch.Tensor] = {}
        self.fact_param_names = [
            "fact.kraw", "fact.nuraw", "fact.th", 
            "fact.L.weight", "fact.R.weight"
        ]
    
    def _compute_fisher_approximations(self, model: NousNet, X_batch: torch.Tensor):
        """Compute diagonal Fisher approximations for BetaFactLayer parameters."""
        if not hasattr(model, "fact") or not hasattr(model.fact, "compute_diff_and_params"):
            return
        
        with torch.no_grad():
            if model.calibrators is not None:
                Xc = torch.stack([cal(xb) for xb, cal in zip(X_batch.t(), model.calibrators)], dim=1)
            else:
                Xc = X_batch
            
            diff, k_vec, nu_vec, net_w = model.fact.compute_diff_and_params(Xc)
            kd = k_vec.unsqueeze(0) * diff
            sig = torch.sigmoid(kd)
            q = 1.0 - sig
            log_sig = F.logsigmoid(kd)
            
            # Fisher for k
            Fk = ((nu_vec.unsqueeze(0) * diff * q) ** 2).mean(dim=0)
            
            # Fisher for nu
            Fnu = (log_sig ** 2).mean(dim=0)
            
            # Fisher for th
            Fth = (((nu_vec * k_vec).unsqueeze(0) * q) ** 2).mean(dim=0)
            
            # Fisher for L and R weights
            x2 = (Xc ** 2).mean(dim=0)
            si = (((nu_vec * k_vec).unsqueeze(0) * q) ** 2).mean(dim=0)
            FL = torch.ger(si, x2)
            FR = FL.clone()
            
            # Update EMA state
            self._update_ema("Fk", Fk)
            self._update_ema("Fnu", Fnu)
            self._update_ema("Fth", Fth)
            self._update_ema("FL", FL)
            self._update_ema("FR", FR)
    
    def _update_ema(self, key: str, new_value: torch.Tensor):
        """Update EMA state with new value."""
        if key not in self.ema_state:
            self.ema_state[key] = new_value.clone()
        else:
            beta = self.defaults['ema_beta']
            self.ema_state[key].mul_(beta).add_(new_value, alpha=1 - beta)
        
        # Add small epsilon for numerical stability
        self.ema_state[key] = self.ema_state[key] + self.defaults['eps']
    
    def _apply_preconditioning(self, model: NousNet):
        """Apply natural gradient preconditioning to gradients."""
        if not hasattr(model, "fact"):
            return
        
        for name, p in model.named_parameters():
            if p.grad is None:
                continue
            
            grad = p.grad.data
            
            # Match parameter names to Fisher approximations
            if name.endswith("fact.kraw") and grad.shape[-1] == self.ema_state["Fk"].shape[0]:
                ng = grad / self.ema_state["Fk"]
            elif name.endswith("fact.nuraw") and grad.shape[-1] == self.ema_state["Fnu"].shape[0]:
                ng = grad / self.ema_state["Fnu"]
            elif name.endswith("fact.th") and grad.shape[-1] == self.ema_state["Fth"].shape[0]:
                ng = grad / self.ema_state["Fth"]
            elif name.endswith("fact.L.weight") and grad.shape == self.ema_state["FL"].shape:
                ng = grad / self.ema_state["FL"]
            elif name.endswith("fact.R.weight") and grad.shape == self.ema_state["FR"].shape:
                ng = grad / self.ema_state["FR"]
            else:
                continue
            
            # Compute trust ratio to prevent exploding updates
            denom = (ng.norm() + 1e-9) / (grad.norm() + 1e-9)
            trust_ratio = torch.clamp(denom, max=self.defaults['trust_clip'])
            
            # Apply preconditioning
            p.grad.data = trust_ratio * ng
    
    def step(self, closure=None, model: Optional[NousNet] = None, X_batch: Optional[torch.Tensor] = None):
        """Performs a single optimization step with natural gradient preconditioning.
        
        Args:
            closure: Optional closure that reevaluates the model and returns the loss
            model: The NousNet model being optimized (required for preconditioning)
            X_batch: Input batch used for computing Fisher approximations
        """
        loss = None
        if closure is not None:
            loss = closure()
        
        if model is None or X_batch is None:
            raise ValueError("BetaNaturalGradient requires model and X_batch arguments")
        
        # Compute Fisher approximations
        self._compute_fisher_approximations(model, X_batch)
        
        # Apply preconditioning to gradients
        self._apply_preconditioning(model)
        
        # Update non-gate parameters with AdamW
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                
                grad = p.grad.data
                state = self.state[p]
                
                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                
                state['step'] += 1
                step = state['step']
                beta1, beta2 = 0.9, 0.999
                lr = group['lr']
                weight_decay = group['weight_decay']
                eps = group['eps']
                
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                
                # Apply weight decay
                if weight_decay != 0:
                    grad = grad.add(p.data, alpha=weight_decay)
                
                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                # Compute bias-corrected moments
                bias_correction1 = 1 - beta1 ** step
                bias_correction2 = 1 - beta2 ** step
                
                # Compute step size
                step_size = lr * np.sqrt(bias_correction2) / bias_correction1
                
                # Update parameters
                denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
                p.data.addcdiv_(exp_avg, denom, value=-step_size)
        
        # Zero gradients after update
        for p in self.gate_params + self.nongate_params:
            if p.grad is not None:
                p.grad.zero_()
        
        return loss


import torch
import pytest
import numpy as np
from nous import NousNet
from nous.optim import (
    MirrorRAdam,
    BetaNaturalGradient,
    ExtragradRAdamTrust,
    SharpnessAwareGates,
    create_optimizer
)

def _create_simple_model():
    """Create a simple model for testing optimizers."""
    return NousNet(
        input_dim=10,
        num_outputs=3,
        task_type="classification",
        num_facts=16,
        rules_per_layer=(8, 4),
        rule_selection_method="soft_fact",
        use_calibrators=False
    )

def _create_test_data(batch_size=32):
    """Create test data for optimization steps."""
    X = torch.randn(batch_size, 10)
    y = torch.randint(0, 3, (batch_size,))
    return X, y

def test_mirror_radam_optimizer():
    """Test MirrorRAdam optimizer basic functionality."""
    model = _create_simple_model()
    X, y = _create_test_data()
    
    optimizer = MirrorRAdam(
        [p for p in model.parameters() if p.requires_grad],
        model=model,
        lr=1e-3,
        lr_gate=1e-2
    )
    
    # Forward pass
    logits = model(X)
    loss = torch.nn.CrossEntropyLoss()(logits, y)
    
    # Backward pass and optimization step
    loss.backward()
    optimizer.step()
    
    # Check that parameters were updated
    param_sum = sum(p.data.sum().item() for p in model.parameters())
    assert not np.isnan(param_sum) and not np.isinf(param_sum)

def test_beta_natural_gradient():
    """Test BetaNaturalGradient optimizer functionality."""
    model = _create_simple_model()
    X, y = _create_test_data()
    
    optimizer = BetaNaturalGradient(
        [p for p in model.parameters() if p.requires_grad],
        model=model,
        lr=1e-3
    )
    
    # Forward pass
    logits = model(X)
    loss = torch.nn.CrossEntropyLoss()(logits, y)
    
    # Backward pass
    loss.backward()
    
    # Optimization step with required model and X_batch arguments
    optimizer.step(model=model, X_batch=X)
    
    # Check that parameters were updated
    param_sum = sum(p.data.sum().item() for p in model.parameters())
    assert not np.isnan(param_sum) and not np.isinf(param_sum)

def test_extragrad_radam_trust():
    """Test ExtragradRAdamTrust optimizer functionality."""
    model = _create_simple_model()
    X, y = _create_test_data()
    
    optimizer = ExtragradRAdamTrust(
        [p for p in model.parameters() if p.requires_grad],
        model=model,
        lr=1e-3
    )
    
    # Define closure for SAM-style optimization
    def closure():
        optimizer.zero_grad()
        logits = model(X)
        loss = torch.nn.CrossEntropyLoss()(logits, y)
        loss.backward()
        return loss
    
    # Optimization step
    optimizer.step(closure=closure, model=model)
    
    # Check that parameters were updated
    param_sum = sum(p.data.sum().item() for p in model.parameters())
    assert not np.isnan(param_sum) and not np.isinf(param_sum)

def test_sharpness_aware_gates():
    """Test SharpnessAwareGates optimizer functionality."""
    model = _create_simple_model()
    X, y = _create_test_data()
    
    optimizer = SharpnessAwareGates(
        [p for p in model.parameters() if p.requires_grad],
        model=model,
        lr=1e-3
    )
    
    # Define closure for SAM-style optimization
    def closure():
        optimizer.zero_grad()
        logits = model(X)
        loss = torch.nn.CrossEntropyLoss()(logits, y)
        loss.backward()
        return loss
    
    # Optimization step
    optimizer.step(closure=closure, model=model)
    
    # Check that parameters were updated
    param_sum = sum(p.data.sum().item() for p in model.parameters())
    assert not np.isnan(param_sum) and not np.isinf(param_sum)

@pytest.mark.parametrize("optimizer_name", [
    "mirror_radam",
    "beta_ng",
    "extragrad_trust",
    "sharpness_aware_gates"
])
def test_optimizer_factory(optimizer_name):
    """Test optimizer factory function."""
    model = _create_simple_model()
    
    optimizer = create_optimizer(
        model,
        optimizer_name,
        lr=1e-3
    )
    
    assert optimizer is not None
    assert hasattr(optimizer, 'step')
    assert hasattr(optimizer, 'zero_grad')

def test_optimizer_parameter_grouping():
    """Test that optimizers correctly group gate and non-gate parameters."""
    model = _create_simple_model()
    
    optimizer = MirrorRAdam(
        [p for p in model.parameters() if p.requires_grad],
        model=model,
        lr=1e-3
    )
    
    # Check that we have both gate and non-gate parameters
    assert len(optimizer.gate_params) > 0
    assert len(optimizer.nongate_params) > 0
    
    # Check parameter categorization
    gate_param_names = []
    nongate_param_names = []
    
    for name, param in model.named_parameters():
        if any(p is param for p in optimizer.gate_params):
            gate_param_names.append(name)
        if any(p is param for p in optimizer.nongate_params):
            nongate_param_names.append(name)
    
    for name in gate_param_names:
        assert ("fact_logits" in name) or ("aggregator_logits" in name) or ("rule_strength_raw" in name)
    
    for name in nongate_param_names:
        assert not (("fact_logits" in name) or ("aggregator_logits" in name) or ("rule_strength_raw" in name))


import numpy as np
import torch
import pytest
from torch.utils.data import TensorDataset, DataLoader

from nous import SoftLogitAND
from nous.training import train_model


def _make_small_binary_dataset(n=512, d=20, seed=0):
    rng = np.random.RandomState(seed)
    X = rng.randn(n, d).astype(np.float32)
    w = rng.randn(d).astype(np.float32)
    logits = (X @ w) * 0.2
    p = 1.0 / (1.0 + np.exp(-logits))
    y = (rng.rand(n) < p).astype(np.float32)
    return X, y


def _make_loaders(X, y, batch_size=64):
    ds = TensorDataset(torch.tensor(X), torch.tensor(y))
    dl = DataLoader(ds, batch_size=batch_size, shuffle=True, drop_last=False)
    return dl


def test_train_model_supports_new_options_sample_average_and_set_to_none():
    X, y = _make_small_binary_dataset(n=512, d=25, seed=1)

    # SoftLogitAND uses init_from_data; feed scaled-like data directly for a smoke test
    model = SoftLogitAND(input_dim=X.shape[1], n_rules=32, n_thresh_per_feat=4, tau=0.7, use_negations=True)
    model.init_from_data(X)

    train_loader = _make_loaders(X[:400], y[:400], batch_size=64)
    val_loader = _make_loaders(X[400:], y[400:], batch_size=64)

    device = torch.device("cpu")
    model.to(device)

    opt = torch.optim.AdamW(model.parameters(), lr=1e-3, weight_decay=1e-4)
    crit = torch.nn.BCEWithLogitsLoss()

    best = train_model(
        model=model,
        train_loader=train_loader,
        val_loader=val_loader,
        criterion=crit,
        optimizer=opt,
        epochs=3,
        patience=2,
        device=device,
        verbose=0,
        print_l0=False,
        clip_grad_max_norm=1.0,
        zero_grad_set_to_none=True,
        loss_average="sample",
    )

    assert isinstance(best, float)
    assert np.isfinite(best)


def test_train_model_rejects_invalid_loss_average():
    X, y = _make_small_binary_dataset(n=128, d=10, seed=2)
    model = SoftLogitAND(input_dim=X.shape[1], n_rules=16, n_thresh_per_feat=4, tau=0.7, use_negations=True)
    model.init_from_data(X)

    dl = _make_loaders(X, y, batch_size=32)
    device = torch.device("cpu")
    model.to(device)

    opt = torch.optim.AdamW(model.parameters(), lr=1e-3)
    crit = torch.nn.BCEWithLogitsLoss()

    with pytest.raises(ValueError):
        train_model(
            model=model,
            train_loader=dl,
            val_loader=dl,
            criterion=crit,
            optimizer=opt,
            epochs=2,
            patience=1,
            device=device,
            verbose=0,
            print_l0=False,
            loss_average="not_a_mode",
        )


from .aggregator import AGG_NAMES, aggregator_mixture_report, format_agg_mixture
from .facts_desc import render_fact_descriptions
from .loo import rule_impact_df
from .mse import minimal_sufficient_explanation
from .pruning import select_pruning_threshold_global, select_pruning_threshold_global_bs
from .global_book import global_rulebook
from .generate import generate_enhanced_explanation
from .fidelity import explanation_fidelity_metrics
from .stability import explanation_stability
from .cf import suggest_rule_counterfactuals

# Explainability for model zoo
from .zoo import (
    describe_threshold_fact,
    softlogitand_global_rules_df,
    softlogitand_local_contrib_df,
    moe_gate_summary_df,
    segmentmoe_local_explain_df,
    hiermoe_local_explain_df,
    nousfamilies_global_summary_df,
    nousfamilies_local_contrib_df,
)

# SoftLogitAND post-hoc (cluster themes + selection strategies)
from .softlogitand_posthoc import (
    SoftLogitANDPosthocReport,
    SoftLogitANDPosthocExplainer,
    mmr_select_rules,
    hard_cap_topk_rules,
    mmr_with_hardcap,
)

__all__ = [
    "AGG_NAMES",
    "aggregator_mixture_report",
    "format_agg_mixture",
    "render_fact_descriptions",
    "rule_impact_df",
    "minimal_sufficient_explanation",
    "select_pruning_threshold_global",
    "select_pruning_threshold_global_bs",
    "global_rulebook",
    "generate_enhanced_explanation",
    "explanation_fidelity_metrics",
    "explanation_stability",
    "suggest_rule_counterfactuals",
    # Zoo explainability
    "describe_threshold_fact",
    "softlogitand_global_rules_df",
    "softlogitand_local_contrib_df",
    "moe_gate_summary_df",
    "segmentmoe_local_explain_df",
    "hiermoe_local_explain_df",
    "nousfamilies_global_summary_df",
    "nousfamilies_local_contrib_df",
    # SoftLogitAND post-hoc
    "SoftLogitANDPosthocReport",
    "SoftLogitANDPosthocExplainer",
    "mmr_select_rules",
    "hard_cap_topk_rules",
    "mmr_with_hardcap",
]


from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
import torch

from sklearn.cluster import AgglomerativeClustering
from sklearn.linear_model import Lasso

from ..zoo import SoftLogitAND
from .zoo import describe_threshold_fact


# ----------------------------
# Numerics / small utilities
# ----------------------------

def _sigmoid_np(x: np.ndarray) -> np.ndarray:
    x = np.asarray(x, dtype=np.float64)
    return 1.0 / (1.0 + np.exp(-x))


def _logit_np(p: np.ndarray, eps: float = 1e-6) -> np.ndarray:
    p = np.clip(np.asarray(p, dtype=np.float64), eps, 1.0 - eps)
    return np.log(p) - np.log1p(-p)


def cosine_sim(a: np.ndarray, b: np.ndarray, eps: float = 1e-12) -> float:
    a = np.asarray(a, dtype=np.float64)
    b = np.asarray(b, dtype=np.float64)
    return float(np.dot(a, b) / ((np.linalg.norm(a) + eps) * (np.linalg.norm(b) + eps)))


def topL_normalized(p: np.ndarray, L: int = 12, eps: float = 1e-12) -> np.ndarray:
    """
    Keep only top-L entries of vector p (by value) and renormalize to sum=1.
    """
    p = np.asarray(p, dtype=np.float64)
    if p.ndim != 1:
        raise ValueError("topL_normalized expects a 1D vector.")
    q = np.zeros_like(p, dtype=np.float64)
    if p.size == 0:
        return q
    L_eff = int(min(max(1, L), p.size))
    idx = np.argsort(-p)[:L_eff]
    q[idx] = p[idx]
    s = float(q.sum())
    if s > 0:
        q /= (s + eps)
    return q


def weighted_quantile(values: np.ndarray, weights: np.ndarray, q: float) -> float:
    """
    Weighted quantile with non-negative weights. If total weight is zero, falls back to median.
    """
    values = np.asarray(values, dtype=np.float64)
    weights = np.asarray(weights, dtype=np.float64)
    if values.size == 0:
        return float("nan")
    order = np.argsort(values)
    v = values[order]
    w = np.clip(weights[order], 0.0, None)
    cw = np.cumsum(w)
    tot = float(cw[-1]) if cw.size else 0.0
    if tot <= 0:
        return float(np.median(values))
    t = float(q) * tot
    idx = int(np.searchsorted(cw, t, side="left"))
    idx = min(max(idx, 0), len(v) - 1)
    return float(v[idx])


# ----------------------------
# SoftLogitAND forward parts
# ----------------------------

@torch.no_grad()
def forward_parts_softlogitand(model: SoftLogitAND, x_scaled_1d: np.ndarray) -> Dict[str, Any]:
    """
    Forward pass returning components needed for post-hoc explanation.

    Returns dict with:
      f: base facts [F0]
      z: rule activations [R]
      P: selector probs [R, Fin]
      w: head weights [R]
      b: head bias (float)
      pred_logit, pred_prob
    """
    model.eval()
    dev = next(model.parameters()).device
    xt = torch.tensor(np.asarray(x_scaled_1d, dtype=np.float32)[None, :], device=dev)

    f = model.facts(xt)     # [1, F0]
    z, P = model.rules(f)   # z: [1, R], P: [R, Fin]
    w = model.head.weight.view(-1)
    b = model.head.bias.view(-1)[0] if model.head.bias is not None else torch.tensor(0.0, device=dev)

    logit = float((z @ w.view(-1, 1) + b).view(-1).item())
    prob = float(_sigmoid_np(np.array([logit], dtype=np.float64))[0])

    return dict(
        f=f.detach().cpu().numpy().squeeze(0),
        z=z.detach().cpu().numpy().squeeze(0),
        P=P.detach().cpu().numpy(),
        w=w.detach().cpu().numpy(),
        b=float(b.detach().cpu().item()),
        pred_logit=logit,
        pred_prob=prob,
    )


@torch.no_grad()
def get_global_P_and_w(model: SoftLogitAND) -> Tuple[np.ndarray, np.ndarray]:
    """
    Return global selector probabilities and head weights:
      P: [R, Fin]
      w: [R]
    """
    model.eval()
    P = model.rules.selector_probs().detach().cpu().numpy()
    w = model.head.weight.view(-1).detach().cpu().numpy()
    return P.astype(np.float64), w.astype(np.float64)


def literal_to_fact_and_neg(model: SoftLogitAND, lit_idx: int) -> Tuple[int, bool]:
    """
    Map literal index in augmented facts to base fact index and negation flag.
    If use_negations=True, Fin = 2*F0, where j>=F0 means negated literal.
    """
    F0 = int(model.facts.num_facts)
    fin = int(model.rules.fin)
    if fin == 2 * F0:
        base = int(lit_idx % F0)
        neg = bool(lit_idx >= F0)
        return base, neg
    return int(lit_idx), False


def literal_text(
    model: SoftLogitAND,
    lit_idx: int,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
) -> str:
    """
    Render a literal as a readable threshold condition.
    """
    base, neg = literal_to_fact_and_neg(model, lit_idx)
    return describe_threshold_fact(model.facts, base, feature_names, scaler=scaler, negated=neg)


def literal_contribs_for_rule(model: SoftLogitAND, f: np.ndarray, P_r: np.ndarray) -> np.ndarray:
    """
    Heuristic local literal score:
      c_j(x) = P_r[j] * logit(fact_aug_j(x))
    This is used only for ranking "top literals" inside a rule.
    """
    F0 = int(model.facts.num_facts)
    fin = int(P_r.shape[0])
    f = np.asarray(f, dtype=np.float64).ravel()

    if fin == 2 * F0:
        f_aug = np.concatenate([f, 1.0 - f], axis=0)
    else:
        f_aug = f

    return np.asarray(P_r, dtype=np.float64) * _logit_np(f_aug)


def rule_feature_mass(model: SoftLogitAND, P_r: np.ndarray, D: int) -> np.ndarray:
    """
    Aggregate selector probability mass from literals to base input features.

    Returns mass vector [D] (normalized to sum=1 if total mass > 0).
    """
    feat_idx = model.facts.feat_idx.detach().cpu().numpy()  # [F0]
    fin = int(P_r.shape[0])
    mass = np.zeros((D,), dtype=np.float64)
    for j in range(fin):
        base, _neg = literal_to_fact_and_neg(model, j)
        feat = int(feat_idx[base])
        if 0 <= feat < D:
            mass[feat] += float(P_r[j])
    s = float(mass.sum())
    if s > 0:
        mass /= s
    return mass


def top_feature_of_rule(model: SoftLogitAND, P_r: np.ndarray, D: int) -> int:
    """
    Dominant feature id of a rule: argmax of feature-mass.
    """
    return int(np.argmax(rule_feature_mass(model, P_r, D)))


def selector_feature_summary(
    model: SoftLogitAND,
    P_vec: np.ndarray,
    feature_names: Sequence[str],
    scaler: Optional[Any] = None,
    top_features: int = 8,
    qlo: float = 0.2,
    qhi: float = 0.8,
) -> List[str]:
    """
    Produce a compact textual summary of where selector mass goes:
      - group by (feature, direction) where direction is "pos" (x > th) or "neg" (x <= th)
      - compute mass and weighted quantile range of thresholds

    Notes:
    - Thresholds are stored in the same space as model inputs (often standardized).
    - If scaler is provided (StandardScaler), thresholds are converted back to raw units.
    """
    feat_idx = model.facts.feat_idx.detach().cpu().numpy()  # [F0]
    th_scaled = model.facts.th.detach().cpu().numpy()       # [F0]
    fin = int(P_vec.shape[0])

    buckets: Dict[Tuple[int, str], Dict[str, list]] = {}
    for j in range(fin):
        p = float(P_vec[j])
        if p <= 0:
            continue

        base, neg = literal_to_fact_and_neg(model, j)
        feat = int(feat_idx[base])
        direction = "neg" if neg else "pos"

        th = float(th_scaled[base])
        if scaler is not None:
            # Invert StandardScaler: raw = scaled * scale + mean
            th = float(th * float(scaler.scale_[feat]) + float(scaler.mean_[feat]))

        key = (feat, direction)
        buckets.setdefault(key, {"p": [], "th": []})
        buckets[key]["p"].append(p)
        buckets[key]["th"].append(th)

    items = []
    for (feat, direction), d in buckets.items():
        p = np.array(d["p"], dtype=np.float64)
        t = np.array(d["th"], dtype=np.float64)
        mass = float(p.sum())
        th_lo = weighted_quantile(t, p, qlo)
        th_hi = weighted_quantile(t, p, qhi)
        th_mid = weighted_quantile(t, p, 0.5)
        items.append((mass, feat, direction, th_lo, th_mid, th_hi))

    items.sort(key=lambda x: -x[0])
    out = []
    for mass, feat, direction, th_lo, th_mid, th_hi in items[: int(top_features)]:
        name = str(feature_names[feat]) if 0 <= feat < len(feature_names) else f"f{feat}"
        if direction == "pos":
            out.append(f"{name} > ~{th_mid:.3g} (range {th_lo:.3g}..{th_hi:.3g}, mass={mass:.2f})")
        else:
            out.append(f"{name} <= ~{th_mid:.3g} (range {th_lo:.3g}..{th_hi:.3g}, mass={mass:.2f})")
    return out


# ----------------------------
# Rule selection strategies
# ----------------------------

def mmr_select_rules(
    contrib_abs: np.ndarray,
    Rep: np.ndarray,
    k: int = 10,
    lambda_div: float = 0.40,
    candidate_pool: int = 140,
) -> List[int]:
    """
    Maximal Marginal Relevance selection:
      score(r) = imp(r) - lambda_div * max_{s in chosen} sim(r, s)
    where imp(r) = |contrib(r)|, sim = cosine similarity in Rep space.
    """
    contrib_abs = np.asarray(contrib_abs, dtype=np.float64).ravel()
    Rep = np.asarray(Rep, dtype=np.float64)
    R = int(contrib_abs.shape[0])
    if R == 0:
        return []

    pool = np.argsort(-contrib_abs)[: min(int(candidate_pool), R)].astype(int).tolist()
    chosen: List[int] = []

    def sim_rs(r: int, s: int) -> float:
        return cosine_sim(Rep[r], Rep[s])

    while len(chosen) < min(int(k), len(pool)):
        best_r = None
        best_score = -1e18
        for r in pool:
            if r in chosen:
                continue
            imp = float(contrib_abs[r])
            if not chosen:
                score = imp
            else:
                max_sim = max(sim_rs(r, s) for s in chosen)
                score = imp - float(lambda_div) * max_sim
            if score > best_score:
                best_score = score
                best_r = r
        if best_r is None:
            break
        chosen.append(int(best_r))
    return chosen


def hard_cap_topk_rules(
    contrib_abs: np.ndarray,
    top_feature_ids: np.ndarray,
    k: int = 10,
    cap_per_feature: int = 2,
) -> List[int]:
    """
    Hardcap strategy:
      - sort rules by importance desc
      - add rule if its dominant feature hasn't exceeded cap_per_feature
    """
    contrib_abs = np.asarray(contrib_abs, dtype=np.float64).ravel()
    top_feature_ids = np.asarray(top_feature_ids, dtype=int).ravel()

    order = np.argsort(-contrib_abs).astype(int).tolist()
    counts: Dict[int, int] = {}
    chosen: List[int] = []
    for r in order:
        f = int(top_feature_ids[r]) if 0 <= r < top_feature_ids.size else -1
        if counts.get(f, 0) >= int(cap_per_feature):
            continue
        chosen.append(int(r))
        counts[f] = counts.get(f, 0) + 1
        if len(chosen) >= int(k):
            break
    return chosen


def mmr_with_hardcap(
    contrib_abs: np.ndarray,
    Rep: np.ndarray,
    top_feature_ids: np.ndarray,
    k: int = 10,
    lambda_div: float = 0.40,
    candidate_pool: int = 140,
    cap_per_feature: int = 2,
) -> List[int]:
    """
    MMR selection with a hard constraint on max rules per dominant feature.
    """
    contrib_abs = np.asarray(contrib_abs, dtype=np.float64).ravel()
    Rep = np.asarray(Rep, dtype=np.float64)
    top_feature_ids = np.asarray(top_feature_ids, dtype=int).ravel()

    R = int(contrib_abs.shape[0])
    if R == 0:
        return []

    pool = np.argsort(-contrib_abs)[: min(int(candidate_pool), R)].astype(int).tolist()
    chosen: List[int] = []
    counts: Dict[int, int] = {}

    def sim_rs(r: int, s: int) -> float:
        return cosine_sim(Rep[r], Rep[s])

    while len(chosen) < min(int(k), len(pool)):
        best_r = None
        best_score = -1e18
        for r in pool:
            if r in chosen:
                continue
            f = int(top_feature_ids[r]) if 0 <= r < top_feature_ids.size else -1
            if counts.get(f, 0) >= int(cap_per_feature):
                continue
            imp = float(contrib_abs[r])
            if not chosen:
                score = imp
            else:
                max_sim = max(sim_rs(r, s) for s in chosen)
                score = imp - float(lambda_div) * max_sim
            if score > best_score:
                best_score = score
                best_r = r

        if best_r is None:
            break
        chosen.append(int(best_r))
        f = int(top_feature_ids[int(best_r)]) if 0 <= int(best_r) < top_feature_ids.size else -1
        counts[f] = counts.get(f, 0) + 1

    return chosen


# ----------------------------
# Markdown rendering (no tabulate dependency)
# ----------------------------

def sanitize_markdown_ascii(s: str) -> str:
    """
    Make markdown output more robust across environments by replacing some unicode dashes/spaces.
    """
    if s is None:
        return ""
    repl = {
        "\u2011": "-", "\u2010": "-", "\u2012": "-", "\u2013": "-", "\u2014": "-", "\u2212": "-",
        "\u00a0": " ",
    }
    for k, v in repl.items():
        s = s.replace(k, v)
    return s


def _md_escape_cell(x: Any) -> str:
    s = sanitize_markdown_ascii(str(x))
    s = s.replace("\t", " ").replace("\n", " ")
    # Robust pipe escaping inside markdown tables
    s = s.replace("|", "&#124;")
    return s


def df_to_markdown_ascii(df: pd.DataFrame, max_rows: int = 12) -> str:
    """
    Simple markdown table renderer without requiring `tabulate`.
    Escapes pipe characters inside cells.
    """
    if df is None or len(df) == 0:
        return "_(empty)_"

    dfx = df.head(int(max_rows)).copy()
    cols = [str(c) for c in dfx.columns.tolist()]

    header = "| " + " | ".join(_md_escape_cell(c) for c in cols) + " |"
    sep = "| " + " | ".join("---" for _ in cols) + " |"

    rows = []
    for _, r in dfx.iterrows():
        rows.append("| " + " | ".join(_md_escape_cell(r[c]) for c in cols) + " |")

    return sanitize_markdown_ascii("\n".join([header, sep] + rows))


# ----------------------------
# Report objects
# ----------------------------

@dataclass
class SoftLogitANDPosthocReport:
    meta: Dict[str, Any]
    tables: Dict[str, pd.DataFrame]
    markdown: str


class SoftLogitANDPosthocExplainer:
    """
    Post-hoc explainer for SoftLogitAND including:
      - multiple rule selection strategies (vanilla/MMR/hardcap/surrogate)
      - cluster themes (rule clustering by selector profile)
      - per-rule dataset statistics (train/test)
      - global importance tables (rules/clusters/features)
      - local markdown report generation

    This class is designed to be fit once on a reference distribution (e.g. train),
    then used for many local explanations.
    """

    def __init__(
        self,
        model: SoftLogitAND,
        feature_names: Sequence[str],
        x_scaler: Optional[Any] = None,
        # Selection / report settings
        k_rules: int = 10,
        k_literals: int = 4,
        summary_top_features: int = 8,
        z_threshold: float = 0.5,
        # MMR settings
        topL: int = 12,
        lambda_div: float = 0.40,
        candidate_pool: int = 140,
        # Hardcap settings
        cap_per_feature: int = 2,
        # Clustering
        n_clusters: int = 16,
        # Surrogate
        use_surrogate: bool = True,
        surrogate_alpha: float = 1e-3,
        # Determinism
        random_state: int = 42,
    ):
        self.model = model
        self.feature_names = list(feature_names)
        self.x_scaler = x_scaler

        self.k_rules = int(k_rules)
        self.k_literals = int(k_literals)
        self.summary_top_features = int(summary_top_features)
        self.z_threshold = float(z_threshold)

        self.topL = int(topL)
        self.lambda_div = float(lambda_div)
        self.candidate_pool = int(candidate_pool)

        self.cap_per_feature = int(cap_per_feature)

        self.n_clusters = int(n_clusters)
        self.use_surrogate = bool(use_surrogate)
        self.surrogate_alpha = float(surrogate_alpha)
        self.random_state = int(random_state)

        # Cached global structures (after fit_posthoc)
        self.P_global: Optional[np.ndarray] = None          # [R, Fin]
        self.w_global: Optional[np.ndarray] = None          # [R]
        self.Rep_topL: Optional[np.ndarray] = None          # [R, Fin]
        self.Rep_featmass: Optional[np.ndarray] = None      # [R, D]
        self.top_feature_ids: Optional[np.ndarray] = None   # [R]

        # Clustering (after fit_posthoc)
        self.cluster_labels: Optional[np.ndarray] = None
        self.cluster_summary: Optional[pd.DataFrame] = None
        self.cluster_Pbar: Optional[Dict[int, np.ndarray]] = None

        # Surrogate (after fit_posthoc)
        self.surrogate: Optional[Lasso] = None
        self.surrogate_coefs: Optional[np.ndarray] = None

        # Dataset stats (after fit_rule_stats)
        self.rule_stats: Optional[pd.DataFrame] = None

    @torch.no_grad()
    def compute_Z_matrix(self, X_scaled: np.ndarray, batch_size: int = 4096) -> np.ndarray:
        """
        Compute rule activation matrix Z of shape [N, R] on scaled inputs.
        """
        self.model.eval()
        dev = next(self.model.parameters()).device
        out = []
        X_scaled = np.asarray(X_scaled, dtype=np.float32)

        for i in range(0, len(X_scaled), int(batch_size)):
            xb = torch.tensor(X_scaled[i:i + int(batch_size)], device=dev)
            f = self.model.facts(xb)
            z, _ = self.model.rules(f)
            out.append(z.detach().cpu().numpy())
        return np.concatenate(out, axis=0).astype(np.float32)

    def fit_posthoc(self, X_ref_scaled: np.ndarray) -> "SoftLogitANDPosthocExplainer":
        """
        Fit global post-hoc structures:
          - selector representations
          - clustering (themes)
          - optional surrogate model
        """
        P, w = get_global_P_and_w(self.model)
        self.P_global = P
        self.w_global = w

        R = int(P.shape[0])
        D = int(len(self.feature_names))

        # Build selector representations
        self.Rep_topL = np.stack([topL_normalized(P[r], L=self.topL) for r in range(R)], axis=0)
        self.Rep_featmass = np.stack([rule_feature_mass(self.model, P[r], D) for r in range(R)], axis=0)
        self.top_feature_ids = np.array([top_feature_of_rule(self.model, P[r], D) for r in range(R)], dtype=int)

        # Cluster rules using cosine distance in top-L selector space
        labels = self._fit_clusters(P=P, w=w, rep=self.Rep_topL)
        self.cluster_labels = labels

        # Fit surrogate: Lasso on Z -> model logits
        if self.use_surrogate:
            Z = self.compute_Z_matrix(X_ref_scaled, batch_size=4096)
            y_teacher_logit = predict_logits_softlogitand(self.model, X_ref_scaled)
            reg = Lasso(alpha=self.surrogate_alpha, max_iter=20000, random_state=self.random_state)
            reg.fit(Z, y_teacher_logit)
            self.surrogate = reg
            self.surrogate_coefs = np.asarray(reg.coef_, dtype=np.float64).ravel()

        return self

    def _fit_clusters(self, P: np.ndarray, w: np.ndarray, rep: np.ndarray) -> np.ndarray:
        """
        Internal clustering routine + cluster summary creation.
        """
        R = int(P.shape[0])
        n_clusters = int(min(max(2, self.n_clusters), max(2, R)))

        # Compute cosine distance matrix in a vectorized way: D = 1 - (X @ X^T)
        X = np.asarray(rep, dtype=np.float64)
        norms = np.linalg.norm(X, axis=1, keepdims=True) + 1e-12
        Xn = X / norms
        sim = Xn @ Xn.T
        Dm = 1.0 - sim
        np.fill_diagonal(Dm, 0.0)

        # sklearn compatibility: "metric" (new) vs "affinity" (old)
        try:
            cl = AgglomerativeClustering(n_clusters=n_clusters, metric="precomputed", linkage="average")
        except TypeError:
            cl = AgglomerativeClustering(n_clusters=n_clusters, affinity="precomputed", linkage="average")

        labels = cl.fit_predict(Dm).astype(int)

        # Build cluster summaries with Pbar weighted by |w|
        self.cluster_Pbar = {}
        rows = []
        for g in sorted(np.unique(labels).tolist()):
            idx = np.where(labels == g)[0].astype(int)
            if idx.size == 0:
                continue
            ww = np.abs(w[idx]) + 1e-12
            Pbar = (P[idx] * ww[:, None]).sum(axis=0)
            Pbar = Pbar / (float(Pbar.sum()) + 1e-12)
            self.cluster_Pbar[int(g)] = Pbar

            rows.append({
                "cluster": int(g),
                "n_rules": int(idx.size),
                "sum_abs_w": float(np.abs(w[idx]).sum()),
                "top_feature_ranges": " || ".join(selector_feature_summary(
                    self.model, Pbar, self.feature_names, scaler=self.x_scaler, top_features=6
                )),
            })

        self.cluster_summary = pd.DataFrame(rows).sort_values("sum_abs_w", ascending=False).reset_index(drop=True)
        return labels

    def fit_rule_stats(
        self,
        X_train_scaled: np.ndarray,
        y_train01: np.ndarray,
        X_test_scaled: np.ndarray,
        y_test01: np.ndarray,
        batch_size: int = 4096,
    ) -> "SoftLogitANDPosthocExplainer":
        """
        Compute per-rule dataset stats for train/test:
          - mean z
          - fraction z > threshold
          - support counts
          - mean y in/out of z>thr
          - delta y (in - out)
          - mean(w*z)
        """
        df_tr = self._compute_rule_stats_on_dataset(
            X_scaled=X_train_scaled,
            y01=y_train01,
            split_name="train",
            z_threshold=self.z_threshold,
            batch_size=batch_size,
        )
        df_te = self._compute_rule_stats_on_dataset(
            X_scaled=X_test_scaled,
            y01=y_test01,
            split_name="test",
            z_threshold=self.z_threshold,
            batch_size=batch_size,
        )
        self.rule_stats = df_tr.merge(df_te, on="rule", how="inner").set_index("rule")
        return self

    @torch.no_grad()
    def add_mean_abs_contrib_to_rule_stats(
        self,
        X_train_scaled: np.ndarray,
        X_test_scaled: np.ndarray,
        batch_size: int = 4096,
    ) -> "SoftLogitANDPosthocExplainer":
        """
        Add mean(abs(w*z)) to rule_stats for train and test splits.
        """
        if self.rule_stats is None:
            raise RuntimeError("Call fit_rule_stats(...) before add_mean_abs_contrib_to_rule_stats(...).")
        self.rule_stats["train_mean_abs_contrib_abs_wz"] = self._mean_abs_contrib_on_dataset(X_train_scaled, batch_size=batch_size)
        self.rule_stats["test_mean_abs_contrib_abs_wz"] = self._mean_abs_contrib_on_dataset(X_test_scaled, batch_size=batch_size)
        return self

    @torch.no_grad()
    def _mean_abs_contrib_on_dataset(self, X_scaled: np.ndarray, batch_size: int = 4096) -> np.ndarray:
        if self.w_global is None:
            raise RuntimeError("Call fit_posthoc(...) first.")
        w = np.asarray(self.w_global, dtype=np.float64)  # [R]
        R = int(w.shape[0])

        self.model.eval()
        dev = next(self.model.parameters()).device
        X_scaled = np.asarray(X_scaled, dtype=np.float32)

        sum_abs = np.zeros((R,), dtype=np.float64)
        n = int(len(X_scaled))
        for i in range(0, n, int(batch_size)):
            xb = torch.tensor(X_scaled[i:i + int(batch_size)], device=dev)
            f = self.model.facts(xb)
            z, _ = self.model.rules(f)
            z = z.detach().cpu().numpy().astype(np.float64)  # [B, R]
            sum_abs += np.abs(z * w[None, :]).sum(axis=0)

        return (sum_abs / max(1, n)).astype(np.float64)

    @torch.no_grad()
    def _compute_rule_stats_on_dataset(
        self,
        X_scaled: np.ndarray,
        y01: np.ndarray,
        split_name: str,
        z_threshold: float,
        batch_size: int = 4096,
    ) -> pd.DataFrame:
        if self.w_global is None:
            raise RuntimeError("Call fit_posthoc(...) first.")
        w = np.asarray(self.w_global, dtype=np.float64)
        R = int(w.shape[0])
        thr = float(z_threshold)

        self.model.eval()
        dev = next(self.model.parameters()).device

        sum_z = np.zeros((R,), dtype=np.float64)
        sum_contrib = np.zeros((R,), dtype=np.float64)

        sum_in = np.zeros((R,), dtype=np.float64)
        sum_y_in = np.zeros((R,), dtype=np.float64)

        y01 = np.asarray(y01, dtype=np.float64).ravel()
        y_total_sum = float(y01.sum())
        X_scaled = np.asarray(X_scaled, dtype=np.float32)
        n = int(len(X_scaled))

        for i in range(0, n, int(batch_size)):
            xb = torch.tensor(X_scaled[i:i + int(batch_size)], device=dev)
            yb = y01[i:i + int(batch_size)]

            f = self.model.facts(xb)
            z, _ = self.model.rules(f)  # [B, R]
            z = z.detach().cpu().numpy().astype(np.float64)

            sum_z += z.sum(axis=0)
            sum_contrib += (z * w[None, :]).sum(axis=0)

            mask = (z > thr).astype(np.float64)
            sum_in += mask.sum(axis=0)
            sum_y_in += mask.T @ yb

        mean_z = sum_z / max(1, n)
        mean_contrib = sum_contrib / max(1, n)
        frac_in = sum_in / max(1, n)

        eps = 1e-12
        mean_y_in = np.full((R,), np.nan, dtype=np.float64)
        mean_y_out = np.full((R,), np.nan, dtype=np.float64)

        in_mask = sum_in > 0
        out_mask = (n - sum_in) > 0

        mean_y_in[in_mask] = sum_y_in[in_mask] / (sum_in[in_mask] + eps)

        sum_out = n - sum_in
        sum_y_out = y_total_sum - sum_y_in
        mean_y_out[out_mask] = sum_y_out[out_mask] / (sum_out[out_mask] + eps)

        df = pd.DataFrame({
            "rule": np.arange(R, dtype=int),
            f"{split_name}_mean_z": mean_z,
            f"{split_name}_frac_z_gt_{thr:.2f}": frac_in,
            f"{split_name}_support_in": sum_in,
            f"{split_name}_mean_y_in_z_gt_thr": mean_y_in,
            f"{split_name}_mean_y_out_z_le_thr": mean_y_out,
            f"{split_name}_delta_y_in_out": mean_y_in - mean_y_out,
            f"{split_name}_mean_contrib_wz": mean_contrib,
        })
        return df

    # ----------------------------
    # Local report tables
    # ----------------------------

    def select_rules_for_x(self, x_raw: np.ndarray, x_is_scaled: bool = False) -> Dict[str, List[int]]:
        """
        Return rule indices for each selection strategy.
        """
        x_s = self._to_scaled_1d(x_raw, x_is_scaled=x_is_scaled)
        parts = forward_parts_softlogitand(self.model, x_s)
        contrib_abs = np.abs(parts["z"] * parts["w"]).astype(np.float64)

        if self.Rep_topL is None or self.Rep_featmass is None or self.top_feature_ids is None:
            raise RuntimeError("Call fit_posthoc(X_ref_scaled) first.")

        vanilla = np.argsort(-contrib_abs)[: self.k_rules].astype(int).tolist()
        mmr_topL = mmr_select_rules(
            contrib_abs=contrib_abs,
            Rep=self.Rep_topL,
            k=self.k_rules,
            lambda_div=self.lambda_div,
            candidate_pool=self.candidate_pool,
        )
        mmr_featmass = mmr_select_rules(
            contrib_abs=contrib_abs,
            Rep=self.Rep_featmass,
            k=self.k_rules,
            lambda_div=self.lambda_div,
            candidate_pool=self.candidate_pool,
        )
        hardcap = hard_cap_topk_rules(
            contrib_abs=contrib_abs,
            top_feature_ids=self.top_feature_ids,
            k=self.k_rules,
            cap_per_feature=self.cap_per_feature,
        )
        mmr_hardcap = mmr_with_hardcap(
            contrib_abs=contrib_abs,
            Rep=self.Rep_topL,
            top_feature_ids=self.top_feature_ids,
            k=self.k_rules,
            lambda_div=self.lambda_div,
            candidate_pool=self.candidate_pool,
            cap_per_feature=self.cap_per_feature,
        )

        out: Dict[str, List[int]] = {
            "vanilla_top_abs_wz": vanilla,
            "mmr_topL": mmr_topL,
            "mmr_feature_mass": mmr_featmass,
            "hardcap_only": hardcap,
            "mmr_plus_hardcap": mmr_hardcap,
        }

        if self.use_surrogate and self.surrogate_coefs is not None:
            sc = np.asarray(self.surrogate_coefs, dtype=np.float64)
            sur_contrib = sc * np.asarray(parts["z"], dtype=np.float64)
            sur_idx = np.argsort(-np.abs(sur_contrib))[: self.k_rules].astype(int).tolist()
            out["surrogate_top_abs_beta_z"] = sur_idx

        return out

    def _rule_table(self, x_scaled_1d: np.ndarray, selected_rules: List[int], title: str = "") -> pd.DataFrame:
        """
        Build a local per-rule table for a given selection list.
        """
        if self.P_global is None or self.w_global is None:
            raise RuntimeError("Call fit_posthoc(...) first.")

        parts = forward_parts_softlogitand(self.model, x_scaled_1d)
        f, z, P_local = parts["f"], parts["z"], parts["P"]
        w = np.asarray(parts["w"], dtype=np.float64)
        contrib = (np.asarray(z, dtype=np.float64) * w).astype(np.float64)

        rows: List[Dict[str, Any]] = []
        for r in selected_rules:
            r = int(r)
            if r < 0 or r >= len(w):
                continue

            P_r = P_local[r]
            summary = " ; ".join(selector_feature_summary(
                self.model, P_r, self.feature_names, scaler=self.x_scaler, top_features=self.summary_top_features
            ))

            # Top literals by heuristic local literal contribution
            c = literal_contribs_for_rule(self.model, f, P_r)
            lit_idx = np.argsort(-np.abs(c))[: self.k_literals].astype(int).tolist()
            lits = [
                f"{literal_text(self.model, j, self.feature_names, scaler=self.x_scaler)} (p={float(P_r[j]):.3f}, c={float(c[j]):+.3f})"
                for j in lit_idx
            ]

            top_feat_id = top_feature_of_rule(self.model, P_r, len(self.feature_names))
            top_feat = self.feature_names[top_feat_id] if 0 <= top_feat_id < len(self.feature_names) else f"f{top_feat_id}"

            row: Dict[str, Any] = {
                "rule": r,
                "top_feature": top_feat,
                "z": float(z[r]),
                "w": float(w[r]),
                "contrib_wz": float(contrib[r]),
                f"z>{self.z_threshold:.2f}?": bool(float(z[r]) > self.z_threshold),
                "rule_summary": summary,
                "top_literals": " | ".join(lits),
            }

            if self.rule_stats is not None and r in self.rule_stats.index:
                st = self.rule_stats.loc[r].to_dict()
                for k, v in st.items():
                    # Keep numeric columns numeric; strings are not expected here
                    try:
                        row[k] = float(v)
                    except Exception:
                        row[k] = v

            rows.append(row)

        df = pd.DataFrame(rows)
        if len(df) > 0:
            df["abs_contrib_tmp"] = df["contrib_wz"].abs()
            df = df.sort_values("abs_contrib_tmp", ascending=False).drop(columns=["abs_contrib_tmp"]).reset_index(drop=True)
        if title:
            df.attrs["title"] = str(title)
        return df

    def _cluster_table(self, x_scaled_1d: np.ndarray, top_clusters: int = 8) -> pd.DataFrame:
        """
        Local cluster theme table with exact contributions (sum of rule contribs in the cluster).
        """
        if self.cluster_labels is None or self.cluster_Pbar is None:
            raise RuntimeError("Call fit_posthoc(...) first (clustering not available).")

        parts = forward_parts_softlogitand(self.model, x_scaled_1d)
        z = np.asarray(parts["z"], dtype=np.float64)
        w = np.asarray(parts["w"], dtype=np.float64)
        P_local = parts["P"]
        contrib = z * w

        rows: List[Dict[str, Any]] = []
        for g in sorted(np.unique(self.cluster_labels).tolist()):
            idx = np.where(self.cluster_labels == g)[0].astype(int)
            if idx.size == 0:
                continue

            cluster_contrib = float(contrib[idx].sum())
            cluster_abs_contrib = float(abs(cluster_contrib))

            top_rule = int(idx[np.argmax(np.abs(contrib[idx]))])

            # Two representative rules by |contrib|
            rep = idx[np.argsort(-np.abs(contrib[idx]))[:2]]
            rep_lines = []
            for r in rep:
                rep_lines.append(
                    f"r{int(r)}: {float(contrib[int(r)]):+.3f} | " +
                    " ; ".join(selector_feature_summary(
                        self.model, P_local[int(r)], self.feature_names, scaler=self.x_scaler, top_features=4
                    ))
                )

            Pbar = self.cluster_Pbar[int(g)]
            rows.append({
                "cluster": int(g),
                "cluster_contrib": cluster_contrib,
                "cluster_abs_contrib": cluster_abs_contrib,
                "n_rules": int(idx.size),
                "top_rule": top_rule,
                "top_rule_contrib": float(contrib[top_rule]),
                "cluster_summary": " || ".join(selector_feature_summary(
                    self.model, Pbar, self.feature_names, scaler=self.x_scaler, top_features=6
                )),
                "representatives": " || ".join(rep_lines),
            })

        df = pd.DataFrame(rows)
        if len(df) == 0:
            return df
        df = df.sort_values("cluster_abs_contrib", ascending=False).reset_index(drop=True)
        return df.head(int(top_clusters))

    def report(
        self,
        x_raw: np.ndarray,
        y_true01: Optional[float] = None,
        x_is_scaled: bool = False,
        top_clusters: int = 8,
    ) -> SoftLogitANDPosthocReport:
        """
        Generate a local report including:
          - multiple rule tables (one per selection strategy)
          - cluster themes table
          - optional surrogate top rules table
          - markdown summary
        """
        x_s = self._to_scaled_1d(x_raw, x_is_scaled=x_is_scaled)
        parts = forward_parts_softlogitand(self.model, x_s)

        pred_prob = float(parts["pred_prob"])
        pred_logit = float(parts["pred_logit"])

        rule_sets = self.select_rules_for_x(x_raw, x_is_scaled=x_is_scaled)

        tables: Dict[str, pd.DataFrame] = {}
        # Always include all strategies if available
        if "mmr_plus_hardcap" in rule_sets:
            tables["mmr_plus_hardcap_rules"] = self._rule_table(x_s, rule_sets["mmr_plus_hardcap"], title="MMR + Hardcap (recommended)")
        if "vanilla_top_abs_wz" in rule_sets:
            tables["vanilla_rules"] = self._rule_table(x_s, rule_sets["vanilla_top_abs_wz"], title="Vanilla top-|w*z|")
        if "mmr_topL" in rule_sets:
            tables["mmr_topL_rules"] = self._rule_table(x_s, rule_sets["mmr_topL"], title="MMR diversified (Top-L selector)")
        if "mmr_feature_mass" in rule_sets:
            tables["mmr_feature_mass_rules"] = self._rule_table(x_s, rule_sets["mmr_feature_mass"], title="MMR diversified (feature-mass)")
        if "hardcap_only" in rule_sets:
            tables["hardcap_rules"] = self._rule_table(x_s, rule_sets["hardcap_only"], title=f"Hardcap (<= {self.cap_per_feature} per dominant feature)")
        if "surrogate_top_abs_beta_z" in rule_sets and self.surrogate_coefs is not None:
            tables["surrogate_top_rules"] = self._surrogate_table(x_s, rule_sets["surrogate_top_abs_beta_z"], title="Surrogate top-|beta*z|")

        if self.cluster_labels is not None:
            tables["cluster_themes"] = self._cluster_table(x_s, top_clusters=int(top_clusters))

        meta: Dict[str, Any] = {
            "pred_prob": pred_prob,
            "pred_logit": pred_logit,
            "y_true01": None if y_true01 is None else float(y_true01),
            "k_rules": self.k_rules,
            "k_literals": self.k_literals,
            "summary_top_features": self.summary_top_features,
            "z_threshold": self.z_threshold,
            "lambda_div": self.lambda_div,
            "topL": self.topL,
            "cap_per_feature": self.cap_per_feature,
            "n_clusters": self.n_clusters,
            "use_surrogate": self.use_surrogate,
            "surrogate_alpha": self.surrogate_alpha,
        }

        md = self.render_markdown(meta, tables)
        return SoftLogitANDPosthocReport(meta=meta, tables=tables, markdown=md)

    def _surrogate_table(self, x_scaled_1d: np.ndarray, selected_rules: List[int], title: str = "") -> pd.DataFrame:
        """
        Surrogate table: contributions based on surrogate coef * z.
        """
        if self.surrogate_coefs is None:
            return pd.DataFrame()

        parts = forward_parts_softlogitand(self.model, x_scaled_1d)
        z = np.asarray(parts["z"], dtype=np.float64)
        P_local = parts["P"]
        sc = np.asarray(self.surrogate_coefs, dtype=np.float64).ravel()
        sur_contrib = sc * z

        rows: List[Dict[str, Any]] = []
        for r in selected_rules:
            r = int(r)
            if r < 0 or r >= sc.size:
                continue
            rows.append({
                "rule": r,
                "coef": float(sc[r]),
                "z": float(z[r]),
                "surrogate_contrib": float(sur_contrib[r]),
                "rule_summary": " ; ".join(selector_feature_summary(
                    self.model, P_local[r], self.feature_names, scaler=self.x_scaler, top_features=self.summary_top_features
                )),
            })

        df = pd.DataFrame(rows)
        if len(df) > 0:
            df["abs_tmp"] = df["surrogate_contrib"].abs()
            df = df.sort_values("abs_tmp", ascending=False).drop(columns=["abs_tmp"]).reset_index(drop=True)
        if title:
            df.attrs["title"] = str(title)
        return df

    def render_markdown(self, meta: Dict[str, Any], tables: Dict[str, pd.DataFrame]) -> str:
        """
        Create a markdown report string (no external dependencies).
        """
        lines: List[str] = []
        lines.append("# SoftLogitAND post-hoc report (binary classification)")
        lines.append("")
        lines.append("## Prediction")
        lines.append(f"- pred_prob: **{meta['pred_prob']:.4f}**")
        lines.append(f"- pred_logit: **{meta['pred_logit']:+.4f}**")
        if meta.get("y_true01", None) is not None:
            y = float(meta["y_true01"])
            lines.append(f"- y_true: **{int(y)}**")
            lines.append(f"- error_prob (pred - y): **{(meta['pred_prob'] - y):+.4f}**")
        lines.append("")
        lines.append("## Settings")
        lines.append(f"- k_rules={meta['k_rules']}, k_literals={meta['k_literals']}")
        lines.append(f"- summary_top_features={meta['summary_top_features']}")
        lines.append(f"- z_threshold={meta['z_threshold']:.2f}")
        lines.append(f"- lambda_div={meta['lambda_div']}, topL={meta['topL']}, cap_per_feature={meta['cap_per_feature']}")
        lines.append(f"- n_clusters={meta['n_clusters']}")
        lines.append(f"- use_surrogate={meta['use_surrogate']} (alpha={meta['surrogate_alpha']})")
        lines.append("")

        order = [
            ("mmr_plus_hardcap_rules", "### A) Recommended: MMR + Hardcap"),
            ("vanilla_rules", "### B) Vanilla top-|w*z|"),
            ("mmr_topL_rules", "### C) MMR diversified (Top-L selector)"),
            ("mmr_feature_mass_rules", "### D) MMR diversified (feature-mass)"),
            ("hardcap_rules", "### E) Hardcap only"),
            ("cluster_themes", "### F) Cluster themes (exact contributions)"),
            ("surrogate_top_rules", "### G) Surrogate top rules (compact governance view)"),
        ]

        for key, title in order:
            if key not in tables:
                continue
            lines.append(sanitize_markdown_ascii(title))
            lines.append(df_to_markdown_ascii(tables[key], max_rows=12))
            lines.append("")

        return sanitize_markdown_ascii("\n".join(lines))

    def save_markdown(self, report: SoftLogitANDPosthocReport, path: str) -> None:
        """
        Save markdown report to file.
        """
        os.makedirs(os.path.dirname(path) or ".", exist_ok=True)
        with open(path, "w", encoding="utf-8", newline="\n") as f:
            f.write(sanitize_markdown_ascii(report.markdown))

    # ----------------------------
    # Fidelity table for strategies
    # ----------------------------

    def fidelity_table_for_x(self, x_raw: np.ndarray, x_is_scaled: bool = False) -> pd.DataFrame:
        """
        Compute simple fidelity metrics for each strategy:
          - coverage_abs_contrib: sum(|w*z| over selected rules) / sum(|w*z| over all rules)
          - avg_pairwise_similarity: average cosine similarity among selected in Rep_topL space
        """
        if self.Rep_topL is None:
            raise RuntimeError("Call fit_posthoc(...) first.")
        x_s = self._to_scaled_1d(x_raw, x_is_scaled=x_is_scaled)
        parts = forward_parts_softlogitand(self.model, x_s)
        contrib_abs = np.abs(np.asarray(parts["z"], dtype=np.float64) * np.asarray(parts["w"], dtype=np.float64))

        rule_sets = self.select_rules_for_x(x_raw, x_is_scaled=x_is_scaled)

        rows = []
        for name, rs in rule_sets.items():
            cov = _coverage_abs_contrib(contrib_abs, rs)
            sim = _avg_pairwise_cosine_similarity(self.Rep_topL[np.array(rs, dtype=int)]) if len(rs) > 1 else 0.0
            rows.append({
                "strategy": name,
                "k_selected": int(len(rs)),
                "coverage_abs_contrib": float(cov),
                "avg_pairwise_similarity": float(sim),
            })

        return pd.DataFrame(rows).sort_values("coverage_abs_contrib", ascending=False).reset_index(drop=True)

    # ----------------------------
    # Global interpretations
    # ----------------------------

    def global_rule_importance(self, split: str = "train", topn: int = 25, sort_by: str = "mean_abs_contrib") -> pd.DataFrame:
        """
        Global rule importance table based on rule_stats.

        sort_by:
          - "mean_abs_contrib" (requires add_mean_abs_contrib_to_rule_stats)
          - "mean_contrib"
          - "coverage"
          - "delta_y"
        """
        if self.rule_stats is None or self.P_global is None or self.w_global is None:
            raise RuntimeError("Need fit_posthoc(...) and fit_rule_stats(...) before global_rule_importance(...).")
        if self.cluster_labels is None or self.top_feature_ids is None:
            raise RuntimeError("Clustering/top_feature_ids missing. Call fit_posthoc(...).")

        rs = self.rule_stats
        thr = self.z_threshold

        col_mean = f"{split}_mean_contrib_wz"
        col_cov = f"{split}_frac_z_gt_{thr:.2f}"
        col_dy = f"{split}_delta_y_in_out"
        col_abs = f"{split}_mean_abs_contrib_abs_wz"

        if sort_by == "mean_contrib":
            key = col_mean
        elif sort_by == "coverage":
            key = col_cov
        elif sort_by == "delta_y":
            key = col_dy
        else:
            # Default: mean_abs_contrib
            if col_abs not in rs.columns:
                raise RuntimeError(f"Column '{col_abs}' not found. Call add_mean_abs_contrib_to_rule_stats(...) first.")
            key = col_abs

        order = rs.sort_values(key, ascending=False).head(int(topn)).index.astype(int).tolist()

        rows = []
        for r in order:
            P_r = self.P_global[int(r)]
            rows.append({
                "rule": int(r),
                "cluster": int(self.cluster_labels[int(r)]),
                "top_feature": self.feature_names[int(self.top_feature_ids[int(r)])],
                "w": float(self.w_global[int(r)]),
                col_cov: float(rs.loc[r, col_cov]),
                col_dy: float(rs.loc[r, col_dy]) if np.isfinite(rs.loc[r, col_dy]) else np.nan,
                col_mean: float(rs.loc[r, col_mean]),
                col_abs: float(rs.loc[r, col_abs]) if col_abs in rs.columns else np.nan,
                "rule_summary": " ; ".join(selector_feature_summary(
                    self.model, P_r, self.feature_names, scaler=self.x_scaler, top_features=8
                )),
            })

        return pd.DataFrame(rows)

    def global_cluster_importance(self, split: str = "train", use_abs: bool = True) -> pd.DataFrame:
        """
        Global cluster importance by summing per-rule metrics inside each cluster.
        Requires rule_stats and clustering.
        """
        if self.cluster_labels is None or self.rule_stats is None or self.cluster_summary is None:
            raise RuntimeError("Need fit_posthoc(...) and fit_rule_stats(...), clustering available.")
        rs = self.rule_stats

        col_mean = f"{split}_mean_contrib_wz"
        col_abs = f"{split}_mean_abs_contrib_abs_wz"
        if col_abs not in rs.columns:
            raise RuntimeError(f"Column '{col_abs}' not found. Call add_mean_abs_contrib_to_rule_stats(...) first.")

        v_mean = rs[col_mean].values.astype(np.float64)
        v_abs = rs[col_abs].values.astype(np.float64)

        rows = []
        for g in sorted(np.unique(self.cluster_labels).tolist()):
            idx = np.where(self.cluster_labels == g)[0]
            rows.append({
                "cluster": int(g),
                "n_rules": int(len(idx)),
                f"{split}_sum_mean_contrib": float(v_mean[idx].sum()),
                f"{split}_sum_mean_abs_contrib": float(v_abs[idx].sum()),
            })

        df = pd.DataFrame(rows)
        key = f"{split}_sum_mean_abs_contrib" if use_abs else f"{split}_sum_mean_contrib"
        df = df.sort_values(key, ascending=False).reset_index(drop=True)

        df = df.merge(
            self.cluster_summary[["cluster", "sum_abs_w", "top_feature_ranges"]],
            on="cluster",
            how="left",
        )
        return df

    def global_feature_importance_mass_weighted(self, split: str = "train", topn: int = 20) -> pd.DataFrame:
        """
        Global per-feature importance computed as:
          signed = sum_r mean(w*z)_r * mass_r(feature)
          abs    = sum_r mean(|w*z|)_r * mass_r(feature)
        """
        if self.rule_stats is None or self.Rep_featmass is None:
            raise RuntimeError("Need fit_posthoc(...) and fit_rule_stats(...).")
        rs = self.rule_stats

        mass = np.asarray(self.Rep_featmass, dtype=np.float64)  # [R, D]
        col_mean = f"{split}_mean_contrib_wz"
        col_abs = f"{split}_mean_abs_contrib_abs_wz"
        if col_abs not in rs.columns:
            raise RuntimeError(f"Column '{col_abs}' not found. Call add_mean_abs_contrib_to_rule_stats(...) first.")

        mean_c = rs[col_mean].values.astype(np.float64)[:, None]
        mean_a = rs[col_abs].values.astype(np.float64)[:, None]

        signed = (mean_c * mass).sum(axis=0)
        absv = (mean_a * mass).sum(axis=0)

        df = pd.DataFrame({
            "feature": self.feature_names,
            f"{split}_mass_weighted_sum": signed,
            f"{split}_mass_weighted_abs_sum": absv,
        }).sort_values(f"{split}_mass_weighted_abs_sum", ascending=False).reset_index(drop=True)

        return df.head(int(topn))

    # ----------------------------
    # Internal helpers
    # ----------------------------

    def _to_scaled_1d(self, x_raw: np.ndarray, x_is_scaled: bool) -> np.ndarray:
        x_raw = np.asarray(x_raw, dtype=np.float32).ravel()
        if x_is_scaled:
            return x_raw
        if self.x_scaler is None:
            raise ValueError("x_scaler must be provided if x_is_scaled=False.")
        return self.x_scaler.transform(x_raw.reshape(1, -1)).astype(np.float32).ravel()


# ----------------------------
# Standalone helper functions
# ----------------------------

@torch.no_grad()
def predict_logits_softlogitand(model: SoftLogitAND, X_scaled: np.ndarray, batch_size: int = 4096) -> np.ndarray:
    """
    Predict logits for SoftLogitAND given scaled inputs X_scaled.
    """
    model.eval()
    dev = next(model.parameters()).device
    X_scaled = np.asarray(X_scaled, dtype=np.float32)
    out = []
    for i in range(0, len(X_scaled), int(batch_size)):
        xb = torch.tensor(X_scaled[i:i + int(batch_size)], device=dev)
        out.append(model(xb).detach().cpu().numpy().ravel())
    return np.concatenate(out, axis=0).astype(np.float64)


def _avg_pairwise_cosine_similarity(mat: np.ndarray) -> float:
    """
    Average pairwise cosine similarity among rows of mat.
    """
    mat = np.asarray(mat, dtype=np.float64)
    k = int(mat.shape[0])
    if k <= 1:
        return 0.0
    norms = np.linalg.norm(mat, axis=1, keepdims=True) + 1e-12
    X = mat / norms
    C = X @ X.T
    iu = np.triu_indices(k, k=1)
    return float(C[iu].mean())


def _coverage_abs_contrib(contrib_abs: np.ndarray, idx: Sequence[int]) -> float:
    """
    Coverage of absolute contributions: sum(abs contrib on selected rules) / sum(abs contrib on all rules).
    """
    contrib_abs = np.asarray(contrib_abs, dtype=np.float64).ravel()
    denom = float(contrib_abs.sum() + 1e-12)
    idx = np.asarray(list(idx), dtype=int)
    if idx.size == 0:
        return 0.0
    return float(contrib_abs[idx].sum() / denom)


from .fixed import FixedPairRuleLayer
from .softmax import SoftmaxRuleLayer
from .sparse import SparseRuleLayer
from .soft_fact import SoftFactRuleLayer
from .blocks import SimpleNousBlock
from .gaters import (
    BaseRuleGater,
    GateSoftRank,
    GateLogisticThresholdExactK,
    GateCappedSimplex,
    GateSparsemaxK,
    GateHardConcreteBudget,
    make_rule_gater,
)

__all__ = [
    "FixedPairRuleLayer",
    "SoftmaxRuleLayer",
    "SparseRuleLayer",
    "SoftFactRuleLayer",
    "SimpleNousBlock",
    # Rule gaters
    "BaseRuleGater",
    "GateSoftRank",
    "GateLogisticThresholdExactK",
    "GateCappedSimplex",
    "GateSparsemaxK",
    "GateHardConcreteBudget",
    "make_rule_gater",
]


from __future__ import annotations
import torch
import torch.nn.functional as F
from torch.optim import Optimizer
import numpy as np
from typing import Dict, Iterable, Optional, Tuple
from .base import BaseOptimizer
from .mirror_radam import MirrorRAdam
from ..model import NousNet

def row_kl_div(p_new: torch.Tensor, p_old: torch.Tensor, eps: float = 1e-8) -> torch.Tensor:
    """Compute KL divergence between two probability distributions row-wise."""
    p_new = torch.clamp(p_new, eps, 1.0)
    p_old = torch.clamp(p_old, eps, 1.0)
    return (p_new * (torch.log(p_new) - torch.log(p_old))).sum(dim=1)

def budget_push_direction(P: torch.Tensor, k_target: int, lam: float = 1e-3) -> torch.Tensor:
    """Compute direction to push towards target budget k_target."""
    B, R = P.shape
    over = P.sum(dim=1).mean() - float(k_target)
    return lam * over * torch.ones_like(P)

class ExtragradRAdamTrust(MirrorRAdam):
    """Extragradient RAdam optimizer with KL trust region for gate parameters.
    Combines extragradient steps with KL trust region constraints for stable
    optimization of rule selection gates.
    Args:
        params: Iterable of parameters to optimize
        model: The model being optimized (required for parameter categorization)
        lr: Learning rate for non-gate parameters (default: 1e-3)
        lr_gate: Learning rate for gate parameters (default: 1e-2)
        betas: Coefficients used for computing running averages of gradient and squared gradient (default: (0.9, 0.999))
        eps: Term added to the denominator to improve numerical stability (default: 1e-8)
        weight_decay: Weight decay coefficient (default: 0)
        delta_kl: Maximum KL divergence allowed between updates (default: 1e-2)
        backtrack: Backtracking factor for KL constraint (default: 0.5)
        lam_budget: Budget constraint coefficient (default: 1e-3)
        k_target: Target number of active rules (default: 8)
        eta_pred: Prediction step size for extragradient (default: 5e-3)
    """
    def __init__(
        self,
        params: Iterable[torch.nn.parameter.Parameter],
        model: torch.nn.Module,
        lr: float = 1e-3,
        lr_gate: float = 1e-2,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0.0,
        delta_kl: float = 1e-2,
        backtrack: float = 0.5,
        lam_budget: float = 1e-3,
        k_target: int = 8,
        eta_pred: float = 5e-3,
    ):
        super().__init__(params, model, lr, lr_gate, betas, eps, weight_decay)
        self.delta_kl = delta_kl
        self.backtrack = backtrack
        self.lam_budget = lam_budget
        self.k_target = k_target
        self.eta_pred = eta_pred

    def step(self, closure=None, model: Optional[NousNet] = None):
        """Performs a single optimization step with extragradient and trust region."""
        loss = None
        if closure is not None:
            loss = closure()
        if model is None:
            model = self.model

        # First, save current gate parameter values
        saved_gate_params = {}
        for name, p in model.named_parameters():
            if p.requires_grad and self._is_gate_param(name) and p.grad is not None and any(p is param for param in self.gate_params):
                saved_gate_params[name] = p.data.clone()

        # First pass: prediction step for extragradient
        for name, p in model.named_parameters():
            if p.requires_grad and self._is_gate_param(name) and p.grad is not None and any(p is param for param in self.gate_params):
                if p.dim() == 2:
                    P = F.softmax(p.data, dim=1)
                    P_tilde = P * torch.exp(-self.eta_pred * p.grad)
                    P_tilde = P_tilde / (P_tilde.sum(dim=1, keepdim=True) + 1e-8)
                    logits_tilde = torch.log(torch.clamp(P_tilde, 1e-8, 1.0))
                    logits_tilde -= logits_tilde.mean(dim=1, keepdim=True)
                    p.data.copy_(logits_tilde)
                else:
                    p.data.add_(-self.eta_pred * p.grad)

        # Second pass: compute gradients at predicted point
        if closure is not None:
            closure()

        # Restore original gate parameters
        for name, p in model.named_parameters():
            if name in saved_gate_params and any(p is param for param in self.gate_params):
                p.data.copy_(saved_gate_params[name])

        # Now perform the actual update with trust region constraints
        super().step(closure)

        # Apply KL trust region constraints to gate parameters
        # betas/eps from group0; lr_gate comes from defaults (gate params aren’t in param_groups)
        group0 = self.param_groups[0] if self.param_groups else {}
        beta1, beta2 = group0.get('betas', (0.9, 0.999))
        eps = group0.get('eps', 1e-8)
        lr_gate = self.defaults.get('lr_gate', group0.get('lr', 1e-2))

        for name, p in model.named_parameters():
            if p.grad is None or not any(p is param for param in self.gate_params) or not self._is_gate_param(name):
                continue

            # Optimizer state keyed by parameter tensor
            state = self.state[p]
            if len(state) == 0:
                state['step'] = 0
                state['exp_avg'] = torch.zeros_like(p.data)
                state['exp_avg_sq'] = torch.zeros_like(p.data)
            state['step'] += 1
            step = state['step']
            grad = p.grad.data

            # Update moving averages
            exp_avg = state['exp_avg']
            exp_avg_sq = state['exp_avg_sq']
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

            # Bias correction
            bias_correction1 = 1 - beta1 ** step
            bias_correction2 = 1 - beta2 ** step
            rt, rectified = self._radam_rectifier(step, beta2)

            # Compute step
            if rectified:
                step_size = lr_gate * rt * np.sqrt(bias_correction2) / bias_correction1
                denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
                step_vec = (exp_avg / denom) * step_size
            else:
                step_size = lr_gate / bias_correction1
                denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
                step_vec = (exp_avg / denom) * step_size

            # Apply trust region constraints
            if p.dim() == 2:
                P_old = F.softmax(p.data, dim=1)
                # Add budget correction
                step_vec = step_vec + budget_push_direction(P_old, k_target=self.k_target, lam=self.lam_budget)
                # Apply backtracking line search with KL constraint
                alpha = 1.0
                while alpha > 1e-4:
                    P_new = P_old * torch.exp(-alpha * step_vec)
                    P_new = P_new / (P_new.sum(dim=1, keepdim=True) + 1e-8)
                    kl = row_kl_div(P_new, P_old)
                    if torch.all(kl <= self.delta_kl):
                        logits_new = torch.log(torch.clamp(P_new, 1e-8, 1.0))
                        logits_new -= logits_new.mean(dim=1, keepdim=True)
                        p.data.copy_(logits_new)
                        break
                    alpha *= self.backtrack
                else:
                    # If backtracking fails, use a small step
                    P_new = P_old * torch.exp(-1e-4 * step_vec)
                    P_new = P_new / (P_new.sum(dim=1, keepdim=True) + 1e-8)
                    logits_new = torch.log(torch.clamp(P_new, 1e-8, 1.0))
                    logits_new -= logits_new.mean(dim=1, keepdim=True)
                    p.data.copy_(logits_new)
            else:
                p.data.add_(-step_vec)

            # Zero the gradient after update
            if p.grad is not None:
                p.grad.zero_()

        return loss


from __future__ import annotations
from typing import Any, Dict, Optional, Union
import torch

from .mirror_radam import MirrorRAdam
from .beta_ng import BetaNaturalGradient
from .extragrad_trust import ExtragradRAdamTrust
from .sam_gates import SharpnessAwareGates

def create_optimizer(
    model: torch.nn.Module, 
    optimizer_name: str,
    lr: float = 1e-3,
    **kwargs: Any
) -> Union[MirrorRAdam, BetaNaturalGradient, ExtragradRAdamTrust, SharpnessAwareGates]:
    """Factory function to create optimizers for Nous models.
    
    Args:
        model: The model to optimize
        optimizer_name: Name of the optimizer to create
            Options: "mirror_radam", "beta_ng", "extragrad_trust", "sharpness_aware_gates"
        lr: Learning rate
        **kwargs: Additional optimizer-specific arguments
    
    Returns:
        Optimizer instance
    """
    params = [p for p in model.parameters() if p.requires_grad]
    
    if optimizer_name.lower() in ["mirror_radam", "radam_kl"]:
        return MirrorRAdam(
            params,
            model=model,
            lr=lr,
            lr_gate=kwargs.get("lr_gate", 1e-2),
            betas=kwargs.get("betas", (0.9, 0.999)),
            eps=kwargs.get("eps", 1e-8),
            weight_decay=kwargs.get("weight_decay", 0.0)
        )
    elif optimizer_name.lower() in ["beta_ng", "betangpp"]:
        return BetaNaturalGradient(
            params,
            model=model,
            lr=lr,
            ema_beta=kwargs.get("ema_beta", 0.9),
            eps=kwargs.get("eps", 1e-6),
            trust_clip=kwargs.get("trust_clip", 10.0),
            weight_decay=kwargs.get("weight_decay", 0.0)
        )
    elif optimizer_name.lower() in ["extragrad_trust", "radam_kl_trust_xtr"]:
        return ExtragradRAdamTrust(
            params,
            model=model,
            lr=lr,
            lr_gate=kwargs.get("lr_gate", 1e-2),
            betas=kwargs.get("betas", (0.9, 0.999)),
            eps=kwargs.get("eps", 1e-8),
            weight_decay=kwargs.get("weight_decay", 0.0),
            delta_kl=kwargs.get("delta_kl", 1e-2),
            backtrack=kwargs.get("backtrack", 0.5),
            lam_budget=kwargs.get("lam_budget", 1e-3),
            k_target=kwargs.get("k_target", 8),
            eta_pred=kwargs.get("eta_pred", 5e-3)
        )
    elif optimizer_name.lower() in ["sharpness_aware_gates", "klsam_gates"]:
        return SharpnessAwareGates(
            params,
            model=model,
            lr=lr,
            rho=kwargs.get("rho", 0.05),
            betas=kwargs.get("betas", (0.9, 0.999)),
            eps=kwargs.get("eps", 1e-8),
            weight_decay=kwargs.get("weight_decay", 0.0),
            clip_grad=kwargs.get("clip_grad", 0.5)
        )
    else:
        raise ValueError(f"Unknown optimizer: {optimizer_name}")


from __future__ import annotations
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim import Optimizer
from typing import Dict, Iterable, Optional, Tuple
import numpy as np

from .base import BaseOptimizer
from ..model import NousNet

class SharpnessAwareGates(BaseOptimizer):
    """Sharpness-Aware Minimization (SAM) optimizer for rule gate parameters.
    
    Applies SAM perturbation only to gate logits in KL geometry (mirror step).
    
    Args:
        params: Iterable of parameters to optimize
        model: The model being optimized (required for parameter categorization)
        lr: Learning rate (default: 1e-3)
        rho: SAM perturbation radius (default: 0.05)
        betas: Coefficients for computing running averages (default: (0.9, 0.999))
        eps: Term added to denominator for numerical stability (default: 1e-8)
        weight_decay: Weight decay coefficient (default: 0)
        clip_grad: Maximum gradient norm for clipping (default: 0.5)
    """
    
    def __init__(
        self,
        params: Iterable[torch.nn.parameter.Parameter],
        model: torch.nn.Module,
        lr: float = 1e-3,
        rho: float = 0.05,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0.0,
        clip_grad: float = 0.5,
    ):
        if lr <= 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if rho <= 0.0:
            raise ValueError(f"Invalid rho value: {rho}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta1 parameter: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta2 parameter: {betas[1]}")
        if eps <= 0.0:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if clip_grad <= 0.0:
            raise ValueError(f"Invalid clip_grad value: {clip_grad}")
        
        defaults = dict(
            lr=lr,
            rho=rho,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            clip_grad=clip_grad,
            step=0
        )
        super().__init__(params, defaults, model=model)
        
        # Create parameter group for non-gate parameters
        self.param_groups = [
            {
                'params': self.nongate_params,
                'lr': lr,
                'betas': betas,
                'eps': eps,
                'weight_decay': weight_decay,
                'clip_grad': clip_grad
            }
        ]
    
    def step(self, closure=None, model: Optional[NousNet] = None):
        """Performs a single SAM optimization step for gate parameters.
        
        Args:
            closure: Optional closure that reevaluates the model and returns the loss
            model: The NousNet model being optimized
        """
        if closure is None:
            raise ValueError("SharpnessAwareGates requires a closure")
        if model is None:
            model = self.model
        
        # Step 1: Store original parameters and compute gradients
        original_params = {}
        for name, p in model.named_parameters():
            if p.requires_grad and self._is_gate_param(name) and any(p is param for param in self.gate_params):
                original_params[name] = p.data.clone()
        
        # First forward-backward pass
        loss = closure()
        nn.utils.clip_grad_norm_(model.parameters(), max_norm=self.defaults['clip_grad'])
        
        # Step 2: Compute SAM perturbation for gate parameters
        with torch.no_grad():
            for name, p in model.named_parameters():
                if name not in original_params or p.grad is None:
                    continue
                
                rho = self.defaults['rho']
                grad = p.grad.data
                
                if p.dim() != 2:
                    # For non-softmax gates
                    pv = p.data.unsqueeze(0)
                    gv = grad.unsqueeze(0)
                    P = F.softmax(pv, dim=1)
                    P_perturbed = P * torch.exp(rho * gv)
                    P_perturbed = P_perturbed / (P_perturbed.sum(dim=1, keepdim=True) + 1e-8)
                    logits = torch.log(torch.clamp(P_perturbed, 1e-8, 1.0))
                    p.data.copy_(logits.squeeze(0))
                else:
                    # For softmax gates (row-wise)
                    P = F.softmax(p.data, dim=1)
                    P_perturbed = P * torch.exp(rho * grad)
                    P_perturbed = P_perturbed / (P_perturbed.sum(dim=1, keepdim=True) + 1e-8)
                    logits = torch.log(torch.clamp(P_perturbed, 1e-8, 1.0))
                    logits -= logits.mean(dim=1, keepdim=True)
                    p.data.copy_(logits)
        
        # Step 3: Second forward-backward pass with perturbed parameters
        loss = closure()
        nn.utils.clip_grad_norm_(model.parameters(), max_norm=self.defaults['clip_grad'])
        
        # Step 4: Restore original parameters and apply update
        for name, p in model.named_parameters():
            if name in original_params and any(p is param for param in self.gate_params):
                p.data.copy_(original_params[name])
        
        # Update non-gate parameters with AdamW
        non_gate_group = self.param_groups[0]
        params = non_gate_group['params']
        lr = non_gate_group['lr']
        weight_decay = non_gate_group['weight_decay']
        beta1, beta2 = non_gate_group['betas']
        eps = non_gate_group['eps']
        clip_grad = non_gate_group['clip_grad']
        step = non_gate_group.get('step', 0) + 1
        non_gate_group['step'] = step
        
        for p in params:
            if p.grad is None:
                continue
            
            grad = p.grad.data
            state = self.state[p]
            
            # State initialization
            if len(state) == 0:
                state['step'] = 0
                state['exp_avg'] = torch.zeros_like(p.data)
                state['exp_avg_sq'] = torch.zeros_like(p.data)
            
            state['step'] += 1
            step = state['step']
            
            exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
            
            # Apply weight decay
            if weight_decay != 0:
                grad = grad.add(p.data, alpha=weight_decay)
            
            # Decay the first and second moment running average coefficient
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
            
            # Compute bias-corrected moments
            bias_correction1 = 1 - beta1 ** step
            bias_correction2 = 1 - beta2 ** step
            
            # Compute step size
            step_size = lr * np.sqrt(bias_correction2) / bias_correction1
            
            # Update parameters
            denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
            p.data.addcdiv_(exp_avg, denom, value=-step_size)
        
        # Update gate parameters with mirror descent
        rho = self.defaults['rho']
        for name, p in model.named_parameters():
            if p.grad is None or not any(p is param for param in self.gate_params) or not self._is_gate_param(name):
                continue
            
            grad = p.grad.data
            if name not in self.state:
                self.state[name] = {}
                self.state[name]['step'] = 0
                self.state[name]['exp_avg'] = torch.zeros_like(p.data)
                self.state[name]['exp_avg_sq'] = torch.zeros_like(p.data)
            
            state = self.state[name]
            state['step'] += 1
            step = state['step']
            
            exp_avg = state['exp_avg']
            exp_avg_sq = state['exp_avg_sq']
            
            # Update moving averages
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
            
            # Bias correction
            bias_correction1 = 1 - beta1 ** step
            bias_correction2 = 1 - beta2 ** step
            
            # Compute step
            step_size = lr / bias_correction1
            denom = (exp_avg_sq.sqrt() / np.sqrt(bias_correction2)).add_(eps)
            step_vec = (exp_avg / denom) * step_size
            
            # Mirror descent step in KL geometry
            if p.dim() == 2:
                P_old = F.softmax(p.data, dim=1)
                P_new = P_old * torch.exp(-step_vec)
                P_new = P_new / (P_new.sum(dim=1, keepdim=True) + 1e-8)
                logits_new = torch.log(torch.clamp(P_new, 1e-8, 1.0))
                logits_new -= logits_new.mean(dim=1, keepdim=True)
                p.data.copy_(logits_new)
            else:
                p.data.add_(-step_vec)
            
            # Zero the gradient after update
            if p.grad is not None:
                p.grad.zero_()
        
        return loss



import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from nous import SoftLogitAND, SegmentMoE, HierarchicalMoE, NousFamilies
from nous.explain import (
    softlogitand_global_rules_df,
    softlogitand_local_contrib_df,
    moe_gate_summary_df,
    segmentmoe_local_explain_df,
    hiermoe_local_explain_df,
    nousfamilies_global_summary_df,
    nousfamilies_local_contrib_df,
)


def test_zoo_explain_outputs_nonempty():
    np.random.seed(0)

    D = 6
    feature_names = ["mean a", "mean b", "worst c", "x error", "other1", "other2"]

    X = np.random.randn(128, D).astype(np.float32)
    x1 = X[0]

    scaler = StandardScaler().fit(X)

    # SoftLogitAND
    m1 = SoftLogitAND(D, n_rules=16, n_thresh_per_feat=4)
    m1.init_from_data(X)
    df_g = softlogitand_global_rules_df(m1, feature_names, scaler=scaler, top_rules=5, top_facts_per_rule=2)
    assert isinstance(df_g, pd.DataFrame) and len(df_g) > 0

    df_l, meta = softlogitand_local_contrib_df(m1, x1, feature_names, scaler=scaler, top_rules=5, top_facts_per_rule=2)
    assert isinstance(df_l, pd.DataFrame) and len(df_l) > 0
    assert "prob_pos" in meta

    # SegmentMoE
    m2 = SegmentMoE(D, feature_names=feature_names, n_segments=4, n_thresh_per_feat=4)
    m2.init_from_data(X)
    df_gate = moe_gate_summary_df(m2, X[:32])
    assert isinstance(df_gate, pd.DataFrame) and len(df_gate) > 0

    df_seg, meta2 = segmentmoe_local_explain_df(m2, x1, feature_names, scaler=scaler, top_segments=2, top_facts=4)
    assert isinstance(df_seg, pd.DataFrame) and len(df_seg) > 0
    assert "prob_pos" in meta2

    # HierarchicalMoE
    m3 = HierarchicalMoE(D, feature_names=feature_names, n_segments=4, n_rules_expert=8, n_thresh_per_feat=4)
    m3.init_from_data(X)
    df_h, meta3 = hiermoe_local_explain_df(
        m3, x1, feature_names, scaler=scaler, top_segments=2, top_rules=3, top_facts_per_rule=2
    )
    assert isinstance(df_h, pd.DataFrame) and len(df_h) > 0
    assert "prob_pos" in meta3

    # NousFamilies
    m4 = NousFamilies(D, feature_names=feature_names, num_facts=16, rules_per_layer=(8, 6), n_families=5)
    m4.init_from_data(X)
    df_fg = nousfamilies_global_summary_df(m4, top_families=3)
    assert isinstance(df_fg, pd.DataFrame) and len(df_fg) > 0

    df_fl, meta4 = nousfamilies_local_contrib_df(m4, x1, top_families=3)
    assert isinstance(df_fl, pd.DataFrame) and len(df_fl) > 0
    assert "prob_pos" in meta4